#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Ironic Networking Network Interface

This is a network interface designed for standalone Ironic deployments that
require minimal network configuration. It implements all required
NetworkInterface methods but performs no actual operations, making it suitable
for environments where external network configuration is handled separately or
not required.
"""

from oslo_config import cfg

import jsonschema
from jsonschema import exceptions as json_schema_exc
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import network
from ironic.common import states
from ironic.conf import ironic_networking
from ironic.drivers import base
from ironic.drivers.modules.network import ironic_networking_schemas
from ironic.drivers.modules.network.switchport_config import SwitchPortConfig
from ironic.networking import api as networking_api

LOG = log.getLogger(__name__)

CONF = cfg.CONF

# Register networking configuration options
ironic_networking.register_opts(CONF)


class IronicNetworking(base.NetworkInterface):
    """Ironic Networking network interface.

    This network interface is designed for standalone Ironic deployments
    where configuration of switch ports should be handled by Ironic.

    Port Configuration:
    This interface validates and processes ports that have an 'extra' property
    containing a 'switchport' sub-property. The switchport configuration must
    conform to the SWITCHPORT_SCHEMA defined in ironic_networking_schemas.py.

    Expected switchport configuration format:
    - mode: 'access', 'trunk', or 'hybrid'
    - native_vlan: VLAN ID (required)
    - allowed_vlans: List of VLAN IDs (required for trunk/hybrid modes only)
    - lag_name: Optional LAG name

    Portgroup Configuration:
    This interface validates and processes portgroups that have an 'extra'
    property containing a 'lag' sub-property. The lag configuration must
    conform to the LAG_SCHEMA defined in ironic_networking_schemas.py.
    """

    def __init__(self):
        """Initialize the IronicNetworking interface."""
        super(IronicNetworking, self).__init__()

        self.switchport_schema = ironic_networking_schemas.SWITCHPORT_SCHEMA
        self.lag_schema = ironic_networking_schemas.LAG_SCHEMA

    def validate(self, task):
        """Validate the network interface configuration.

        For the ironic networking interface, this validates any ports
        that have switchport configuration in their 'extra' field and any
        portgroups that have LAG configuration in their 'extra' field.

        :param task: A TaskManager instance.
        :raises: InvalidParameterValue if switchport or lag
                 configuration is invalid.
        """
        # Validate ports with switchport configuration
        for port in task.ports:
            self._validate_port_switchport_config(task, port)

        # Validate portgroups with LAG configuration
        for portgroup in task.portgroups:
            self._validate_portgroup_lag_config(portgroup)
            self._validate_portgroup_member_ports(task, portgroup)

    @staticmethod
    def _get_switch_port_config(task, network_type):
        """Get the SwitchPortConfig for a given network type.

        The value is determined by first checking the node's driver_info
        for an override, and then falling back to the global conf option.

        :param task: A TaskManager instance.
        :param network_type: One of 'cleaning', 'provisioning', 'servicing',
                             'rescuing', or 'inspection'.
        :returns: A SwitchPortConfig instance, or None if not set.
        :raises: InvalidParameterValue if the network value is set but
                 has an invalid format.
        """
        node = task.node
        driver_info_key = f'{network_type}_network'
        conf_key = f'{network_type}_network'

        # Check driver_info override
        network_value = node.driver_info.get(driver_info_key)
        if not network_value:
            # Fallback to global conf
            try:
                network_value = getattr(CONF.ironic_networking, conf_key, '')
                LOG.debug("Retrieved config value for %s: %s",
                          conf_key, network_value)
            except AttributeError as e:
                LOG.warning(
                    "Networking configuration not available or option '%s' "
                    "not found: %s", conf_key, e)
                network_value = ''

        if not network_value:
            LOG.warning(
                "No network configured for network type '%s'. "
                "Falling back to port's switchport attributes.",
                network_type
            )
            return None

        config = SwitchPortConfig.from_string(network_value, network_type)

        LOG.debug(
            "Called _get_switch_port_config with node=%(node)s, "
            "network_type=%(network_type)s; returning %(config)s "
            "from value='%(value)s'",
            {'node': task.node.uuid, 'network_type': network_type,
             'config': config, 'value': network_value})

        return config

    def _validate_network_requirements(self, task, network_type):
        """Validate that at least one port has required network configuration.

        This method checks that at least one port on the node has both
        local_link_connection information and valid mode/native_vlan
        configuration either from the network configuration or the port's
        switchport configuration.

        :param task: A TaskManager instance.
        :param network_type: The type of network to validate (e.g.,
                             'cleaning', 'rescuing', 'inspection',
                             'servicing').
        :raises: InvalidParameterValue if unable to parse network configuration
        """
        for port in task.ports:
            # Check if port has local_link_connection
            if not port.local_link_connection:
                continue

            # Try to get switchport info from port.extra
            switchport = port.extra.get('switchport') if port.extra else None
            if not switchport:
                continue

            # Get the SwitchPortConfig for the network type
            config = self._get_switch_port_config(task, network_type)

            # If network configuration is not set, use port's switchport info
            if config is None or not config.is_valid:
                config = SwitchPortConfig.from_switchport(switchport)

            # If we have valid configuration, we found at least one valid port
            if config and config.is_valid:
                return

        # TODO(alegacy): Need to consider how far the validation should go.
        # For example, the initial inspection may not have any ports so no
        # validation is required, but then later we could cause issues with
        # cleaning or rescuing if we require ports to be present and
        # configured.  Needs more thought.  Maybe just enforce that ports
        # that have switchport configuration must have a
        # local_link_connection?
        LOG.warning(
            _("Node %(node)s requires at least one port with "
              "local_link_connection and valid mode/native_vlan "
              "configuration for %(network_type)s network"),
            {'node': task.node.uuid, 'network_type': network_type})

    def _validate_port_switchport_config(self, task, port):
        """Validate switchport configuration in port's extra field.

        :param task: A TaskManager instance.
        :param port: A Port object.
        :raises: InvalidParameterValue if switchport configuration is invalid.
        """
        if not port.extra:
            return

        switchport_config = port.extra.get('switchport')
        if not switchport_config:
            return

        try:
            jsonschema.validate(switchport_config, self.switchport_schema)
        except json_schema_exc.ValidationError as e:
            raise exception.InvalidParameterValue(
                _("Invalid switchport configuration for port %(port)s: "
                  "%(error)s") % {'port': port.uuid, 'error': e})

        # Validate that PXE-enabled ports use access mode
        if port.pxe_enabled:
            mode = switchport_config.get('mode')
            if mode != 'access':
                raise exception.InvalidParameterValue(
                    _("Port %(port)s is PXE-enabled but has switchport mode "
                      "'%(mode)s'. PXE-enabled ports must use 'access' mode "
                      "when switchport configuration is present.")
                    % {'port': port.uuid, 'mode': mode})

        # Validate that local_link_connection has required fields
        if not port.local_link_connection:
            # Inspection is the only state where not having a
            # local_link_connection is permitted since we expect it to be
            # populated by LLDP if it is available.  If the intent is not to
            # use inspection then the user should have manually populated
            # this field.
            if task.node.provision_state not in states.INSPECTION_STATES:
                raise exception.InvalidParameterValue(
                    _("Port %(port)s has switchport configuration but is "
                      "missing local_link_connection in state: %(state)s.")
                    % {'port': port.uuid, 'state': task.node.provision_state})

            # During inspection, we allow this and just log it hoping that
            # inspection will populate it.
            LOG.debug(
                "Port %s is missing local_link_connection in state: %s",
                port.uuid, task.node.provision_state)
            return

        switch_id = port.local_link_connection.get('switch_id')
        port_id = port.local_link_connection.get('port_id')

        errors = []
        if not switch_id:
            errors.append("'switch_id'")
        if not port_id:
            errors.append("'port_id'")

        if errors:
            raise exception.InvalidParameterValue(
                _("Port %(port)s local_link_connection missing required "
                  "field(s): %(fields)s")
                % {'port': port.uuid, 'fields': ', '.join(errors)})

    def _validate_portgroup_lag_config(self, portgroup):
        """Validate LAG configuration in portgroup's extra field.

        :param portgroup: A Portgroup object.
        :raises: InvalidParameterValue if LAG configuration is
                 invalid.
        """
        if not portgroup.extra:
            return

        lag_config = portgroup.extra.get('lag')
        if not lag_config:
            return

        try:
            jsonschema.validate(lag_config, self.lag_schema)
            LOG.debug("Portgroup %s LAG configuration is valid",
                      portgroup.uuid)
        except json_schema_exc.ValidationError as e:
            raise exception.InvalidParameterValue(
                _("Invalid LAG configuration for portgroup "
                  "%(portgroup)s: %(error)s")
                % {'portgroup': portgroup.uuid, 'error': e})

    @staticmethod
    def _validate_portgroup_member_ports(task, portgroup):
        """Validate that portgroup member ports have switchport configuration.

        :param task: A TaskManager instance.
        :param portgroup: A Portgroup object.
        :raises: InvalidParameterValue if member ports lack switchport
                 configuration.
        """
        if not portgroup.extra or not portgroup.extra.get('lag'):
            return

        # Find member ports of this portgroup
        member_ports = [port for port in task.ports
                        if port.portgroup_id == portgroup.id]

        if not member_ports:
            raise exception.InvalidParameterValue(
                _("Portgroup %(portgroup)s has LAG configuration "
                  "but no member ports") % {'portgroup': portgroup.uuid})

        # Check that all member ports have switchport configuration
        ports_without_switchport = []
        ports_without_local_link = []
        for port in member_ports:
            if not port.extra or not port.extra.get('switchport'):
                ports_without_switchport.append(port.uuid)
            elif (not port.local_link_connection
                  or not port.local_link_connection.get('switch_id')
                  or not port.local_link_connection.get('port_id')):
                ports_without_local_link.append(port.uuid)

        errors = []
        if ports_without_switchport:
            errors.append(
                _("member ports %(ports)s lack switchport configuration")
                % {'ports': ', '.join(ports_without_switchport)})
        if ports_without_local_link:
            errors.append(
                _("member ports %(ports)s lack proper local_link_connection "
                  "with switch_id and port_id fields")
                % {'ports': ', '.join(ports_without_local_link)})

        if errors:
            raise exception.InvalidParameterValue(
                _("Portgroup %(portgroup)s has LAG configuration but: "
                  "%(errors)s")
                % {'portgroup': portgroup.uuid, 'errors': '; '.join(errors)})

    @staticmethod
    def _get_portgroup_switch_ids(task, portgroup):
        """Get switch IDs from member ports of a portgroup.

        :param task: A TaskManager instance.
        :param portgroup: A Portgroup object.
        :returns: List of unique switch IDs from member ports, or None if
                  no valid ports.
        """
        if not portgroup.extra or not portgroup.extra.get('lag'):
            return None

        # Find member ports of this portgroup
        member_ports = [port for port in task.ports
                        if port.portgroup_id == portgroup.id]

        if not member_ports:
            return None

        # Extract switch IDs from member ports' local_link_connection
        switch_ids = set()
        for port in member_ports:
            if (port.extra and port.extra.get('switchport')
                    and port.local_link_connection):
                switch_id = port.local_link_connection.get('switch_id')
                if switch_id:
                    switch_ids.add(switch_id)

        return list(switch_ids) if switch_ids else None

    @staticmethod
    def _get_portgroup_lag_name(portgroup):
        """Get LAG name from portgroup configuration.

        For now, we'll use the portgroup name as the LAG name.
        This could be enhanced to use a specific field in the future.

        :param portgroup: A Portgroup object.
        :returns: LAG name string.
        """
        if portgroup.name:
            return portgroup.name
        else:
            # TODO(alegacy): naming of LAG instances needs to be re-visited
            # when that part of the feature is completed.  Some switches will
            # want to name the LAG itself and others may have specific naming
            # requirements which may make this tricky at best.
            return f"lag-{portgroup.uuid[:8]}"

    @staticmethod
    def _get_port_switch_info(port):
        """Get switch ID and port name from port's local_link_connection.

        :param port: A Port object.
        :returns: Tuple of (switch_id, port_name) or (None, None) if not
                  available.
        """
        if not port.local_link_connection:
            return None, None

        switch_id = port.local_link_connection.get('switch_id')
        port_name = port.local_link_connection.get('port_id')

        return switch_id, port_name

    @staticmethod
    def _get_port_description(port):
        """Generate description for a port.

        :param port: A Port object.
        :returns: Description string.
        """
        return f"Ironic Port {port.uuid}"

    @staticmethod
    def _get_portgroup_description(portgroup):
        """Generate description for a portgroup.

        :param portgroup: A Portgroup object.
        :returns: Description string.
        """
        return f"Ironic PortGroup {portgroup.uuid}"

    def _resolve_network_configuration(self, task, port_obj, network_type):
        """Resolve network configuration for a port based on network type.

        :param task: A TaskManager instance.
        :param port_obj: A Port object.
        :param network_type: The type of network to resolve configuration for.
        :returns: A SwitchPortConfig instance, or None if not configured.
        :raises: InvalidParameterValue if unable to parse network
            configuration
        """
        # Try network-specific configuration first
        config = self._get_switch_port_config(task, network_type)

        if config is None or not config.is_valid:
            # Fallback to port's switchport configuration unless the network
            # is the idle network in which case we simply allow the port to
            # remain unconfigured (or configured to the switch-wide default)
            if network_type == network.IDLE_NETWORK:
                return None
            switchport = (port_obj.extra.get('switchport', {})
                          if port_obj.extra else {})
            config = SwitchPortConfig.from_switchport(switchport)

        LOG.debug(
            "Resolving network configuration for port %(port)s, "
            "network_type=%(network_type)s: %(config)s",
            {
                'port': getattr(port_obj, 'uuid', None),
                'network_type': network_type,
                'config': config,
            })
        return config

    def _get_original_port_config(self, task, port_obj):
        """Get original port configuration before changes.

        Called from port_changed() to retrieve the port's configuration
        from task.ports before the current changes were applied. This
        allows comparison between original and new values to determine
        what network operations are needed.

        :param task: A TaskManager instance.
        :param port_obj: The changed Port object with new values.
        :returns: Tuple of (original_port, original_switchport,
            original_local_link) where original_port is the Port object
            from task.ports, original_switchport is the switchport dict
            from port.extra, and original_local_link is the
            local_link_connection dict. Any value may be None if not found.
        """
        original_port = None
        for port in task.ports:
            if port.uuid == port_obj.uuid:
                original_port = port
                break

        original_switchport = None
        original_local_link = None
        if original_port:
            original_switchport = (original_port.extra.get('switchport')
                                   if original_port.extra else None)
            original_local_link = original_port.local_link_connection

        return original_port, original_switchport, original_local_link

    def _determine_port_changes(self, original_switchport, current_switchport,
                                original_local_link, current_local_link):
        """Determine what changes occurred to the port.

        Called from port_changed() after retrieving original configuration
        to classify the type of change that occurred. The result is used
        to decide which network operations (reset, update, etc.) are needed.

        :param original_switchport: The switchport dict from the original
            port.extra, or None.
        :param current_switchport: The switchport dict from the changed
            port.extra, or None.
        :param original_local_link: The original local_link_connection
            dict, or None.
        :param current_local_link: The current local_link_connection
            dict, or None.
        :returns: Tuple of (switchport_removed, local_link_removed,
            local_link_changed, switchport_changed, switchport_added)
            where each value is a boolean indicating that type of change.
        """
        switchport_removed = original_switchport and not current_switchport
        local_link_removed = original_local_link and not current_local_link
        local_link_changed = (original_local_link != current_local_link
                              and original_local_link
                              and current_local_link)
        switchport_changed = (original_switchport != current_switchport
                              and original_switchport
                              and current_switchport)
        switchport_added = not original_switchport and current_switchport

        return (switchport_removed, local_link_removed, local_link_changed,
                switchport_changed, switchport_added)

    def _handle_port_removal(self, task, port_obj, original_port,
                             original_local_link, original_switchport,
                             active_network_type):
        """Handle port removal cases.

        Called when switchport or local_link_connection is removed from a port.
        Resets the switch port configuration using the original values.

        :param task: A TaskManager instance.
        :param port_obj: The changed Port object.
        :param original_port: The original Port object from task.ports.
        :param original_local_link: The original local_link_connection dict.
        :param original_switchport: The original switchport configuration.
        :param active_network_type: The network type that should be active.
        :raises: InvalidParameterValue if unable to parse network configuration
        :raises: NetworkError if the networking service call fails
        """
        LOG.debug(
            "Port %(port)s switchport or local_link_connection removed",
            {'port': port_obj.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        if original_local_link and original_switchport:
            # Get configuration for reset operation using original values
            original_config = self._resolve_network_configuration(
                task, original_port, active_network_type)

            if original_config and original_config.is_valid:
                # Get switch and port info from original state
                switch_id = original_local_link.get('switch_id')
                port_name = original_local_link.get('port_id')

                if switch_id and port_name:
                    idle_vlan = (idle_config.native_vlan
                                 if idle_config else None)
                    result = networking_api.reset_port(
                        task.context, switch_id, port_name,
                        original_config.native_vlan,
                        allowed_vlans=original_config.allowed_vlans,
                        default_vlan=idle_vlan)
                    LOG.debug(
                        "Successfully reset port %(port_name)s on "
                        "switch %(switch_id)s via networking service: "
                        "%(result)s",
                        {'port_name': port_name,
                         'switch_id': switch_id,
                         'result': result})
                else:
                    LOG.warning(
                        "Cannot reset port %(port)s: missing "
                        "switch_id or port_id",
                        {'port': port_obj.uuid})
            elif active_network_type != 'idle':
                LOG.warning(
                    "Cannot reset port %(port)s: missing mode or "
                    "native_vlan for %(active_network_type)s network",
                    {'port': port_obj.uuid})

    def _handle_local_link_change(self, task, port_obj, original_port,
                                  original_local_link, original_switchport,
                                  active_network_type):
        """Handle local_link_connection change (requires reset + update).

        Called when the local_link_connection changes on a port. Resets the
        old switch port configuration before the new configuration is applied.

        :param task: A TaskManager instance.
        :param port_obj: The changed Port object.
        :param original_port: The original Port object from task.ports.
        :param original_local_link: The original local_link_connection dict.
        :param original_switchport: The original switchport configuration.
        :param active_network_type: The network type that should be active.
        :raises: InvalidParameterValue if unable to parse network configuration
        :raises: NetworkError if the networking service call fails
        """
        LOG.debug("Port %(port)s local_link_connection changed",
                  {'port': port_obj.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        # First, reset the old port
        if original_local_link and original_switchport:
            original_config = self._resolve_network_configuration(
                task, original_port, active_network_type)

            if original_config and original_config.is_valid:
                switch_id = original_local_link.get('switch_id')
                port_name = original_local_link.get('port_id')

                if switch_id and port_name:
                    idle_vlan = (idle_config.native_vlan
                                 if idle_config else None)
                    result = networking_api.reset_port(
                        task.context, switch_id, port_name,
                        original_config.native_vlan,
                        allowed_vlans=original_config.allowed_vlans,
                        default_vlan=idle_vlan)
                    LOG.debug(
                        "Successfully reset old port %(port_name)s "
                        "on switch %(switch_id)s via networking "
                        "service: %(result)s",
                        {'port_name': port_name,
                         'switch_id': switch_id,
                         'result': result})

    def _handle_switchport_update(self, task, port_obj, current_local_link,
                                  current_switchport, active_network_type,
                                  switchport_added):
        """Handle switchport addition or modification.

        Called when switchport configuration is added or changed on a port.
        Validates the configuration and updates the switch port.

        :param task: A TaskManager instance.
        :param port_obj: The changed Port object.
        :param current_local_link: The current local_link_connection dict.
        :param current_switchport: The current switchport configuration.
        :param active_network_type: The network type that should be active.
        :param switchport_added: True if switchport was added, False if
            updated.
        :raises: InvalidParameterValue if unable to parse network configuration
        :raises: NetworkError if the networking service call fails
        """
        if switchport_added:
            LOG.info("Port %(port)s switchport configuration added",
                     {'port': port_obj.uuid})
        else:
            LOG.info("Port %(port)s switchport configuration updated",
                     {'port': port_obj.uuid})

        if not current_local_link:
            return

        # Validate the current switchport configuration
        self._validate_port_switchport_config(task, port_obj)

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        # Resolve configuration based on active network type
        config = self._resolve_network_configuration(
            task, port_obj, active_network_type)

        if config and config.is_valid:
            # Get switch and port info from current state
            switch_id = current_local_link.get('switch_id')
            port_name = current_local_link.get('port_id')
            description = self._get_port_description(port_obj)
            lag_name = current_switchport.get('lag_name')

            if switch_id and port_name:
                idle_vlan = (idle_config.native_vlan
                             if idle_config else None)
                result = networking_api.update_port(
                    task.context, switch_id, port_name,
                    description, config.mode, config.native_vlan,
                    allowed_vlans=config.allowed_vlans,
                    lag_name=lag_name,
                    default_vlan=idle_vlan)
                LOG.debug(
                    "Successfully updated port %(port_name)s "
                    "on switch %(switch_id)s for %(network_type)s "
                    "network via networking service: %(result)s",
                    {'port_name': port_name,
                     'switch_id': switch_id,
                     'network_type': active_network_type,
                     'result': result})
            else:
                LOG.warning(
                    "Cannot update port %(port)s: missing "
                    "switch_id or port_id",
                    {'port': port_obj.uuid})
        else:
            LOG.warning(
                "Cannot update port %(port)s: missing mode or "
                "native_vlan for %(network_type)s network",
                {'port': port_obj.uuid,
                 'network_type': active_network_type})

    def port_changed(self, task, port_obj):
        """Handle any actions required when a port changes.

        In the ironic networking interface, this method processes
        ports that have switchport configuration in their 'extra' field
        and calls the networking service API accordingly based on the
        node's current provision state and what network should be active.

        :param task: A TaskManager instance.
        :param port_obj: A changed Port object.
        """
        # Check if relevant fields have changed
        changed_fields = port_obj.obj_what_changed()
        if not ({'extra', 'local_link_connection'} & changed_fields):
            # No relevant changes, skip this port
            return

        # Determine what network should be active based on node state
        active_network_type = network.get_network_type_for_state(
            task.node.provision_state)

        LOG.debug(
            "Port %(port)s changed on node %(node)s in state %(state)s, "
            "active network type: %(network_type)s, changed fields: "
            "%(fields)s",
            {'port': port_obj.uuid, 'node': task.node.uuid,
             'state': task.node.provision_state,
             'network_type': active_network_type,
             'fields': list(changed_fields)})

        # Get current configuration
        current_switchport = (port_obj.extra.get('switchport')
                              if port_obj.extra else None)
        current_local_link = port_obj.local_link_connection

        # Get original port configuration (before changes)
        (original_port, original_switchport,
         original_local_link) = self._get_original_port_config(task, port_obj)

        # Determine what actions need to be taken
        (switchport_removed, local_link_removed, local_link_changed,
         switchport_changed, switchport_added) = self._determine_port_changes(
            original_switchport, current_switchport,
            original_local_link, current_local_link)

        # Handle removal cases
        if switchport_removed or local_link_removed:
            self._handle_port_removal(task, port_obj, original_port,
                                      original_local_link, original_switchport,
                                      active_network_type)
            return

        # Handle local_link_connection change (requires reset + update)
        if local_link_changed:
            self._handle_local_link_change(task, port_obj, original_port,
                                           original_local_link,
                                           original_switchport,
                                           active_network_type)
            # Then configure the new port (fall through to update logic)
            switchport_changed = True  # Force update with new configuration

        # Handle switchport addition or modification
        if switchport_added or switchport_changed:
            self._handle_switchport_update(task, port_obj, current_local_link,
                                           current_switchport,
                                           active_network_type,
                                           switchport_added)

    def portgroup_changed(self, task, portgroup_obj):
        """Handle any actions required when a portgroup changes.

        In the ironic networking interface, portgroup changes are not
        currently supported and will be logged but no action will be taken.

        :param task: A TaskManager instance.
        :param portgroup_obj: A changed Portgroup object.
        """
        LOG.debug(
            "Portgroup %(portgroup)s (%(name)s) configuration changed - "
            "portgroup changes not currently supported by "
            "ironic-networking interface",
            {'portgroup': portgroup_obj.uuid,
             'name': portgroup_obj.name or 'unnamed'})

    def vif_attach(self, task, vif_info):
        """Attach a virtual network interface to a node.

        In the ironic networking interface, VIF attachment is a no-op.
        This allows the operation to complete successfully without performing
        any actual network configuration.

        :param task: A TaskManager instance.
        :param vif_info: A dictionary of information about a VIF.
            It must have an 'id' key, whose value is a unique
            identifier for that VIF.
        """
        pass

    def vif_detach(self, task, vif_id):
        """Detach a virtual network interface from a node.

        In the ironic networking interface, VIF detachment is a no-op.
        This allows the operation to complete successfully without performing
        any actual network configuration.

        :param task: A TaskManager instance.
        :param vif_id: A VIF ID to detach.
        """
        pass

    def vif_list(self, task):
        """List attached VIF IDs for a node.

        In the ironic networking interface, no VIFs are ever attached,
        so this always returns an empty list.

        :param task: A TaskManager instance.
        :returns: Empty list as no VIFs are managed by this interface.
        """
        return []

    def get_current_vif(self, task, p_obj):
        """Return the currently used VIF associated with port or portgroup.

        In the ironic networking interface, no VIFs are managed,
        so this always returns None.

        :param task: A TaskManager instance.
        :param p_obj: Ironic port or portgroup object.
        :returns: None as no VIFs are managed by this interface.
        """
        return None

    def _add_tenant_networks(self, task):
        """Add tenant networks to a node.

        This method attempts to configure the tenant network for each port
        on the node using only the port's switchport configuration. If the
        switchport configuration is missing or incomplete, the port is skipped.

        :param task: A TaskManager instance.
        """
        LOG.debug("Adding tenant networks to node %(node)s",
                  {'node': task.node.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        for port in task.ports:
            switchport = port.extra.get('switchport') if port.extra else None
            link_info = port.local_link_connection

            if not switchport or not link_info:
                LOG.debug(
                    "Skipping port %(port)s: missing switchport or "
                    "local_link_connection info for tenant network setup.",
                    {'port': port.uuid}
                )
                continue

            config = SwitchPortConfig.from_switchport(switchport)

            if not config or not config.is_valid:
                LOG.debug(
                    "Skipping port %(port)s: missing mode or native_vlan or "
                    "allowed_vlans for tenant network setup.",
                    {'port': port.uuid}
                )
                continue

            try:
                idle_vlan = (idle_config.native_vlan
                             if idle_config else None)
                networking_api.update_port(
                    task.context,
                    link_info.get('switch_id'),
                    link_info.get('port_id'),
                    self._get_port_description(port),
                    config.mode,
                    config.native_vlan,
                    allowed_vlans=config.allowed_vlans,
                    lag_name=None,
                    default_vlan=idle_vlan,
                )
            except (exception.InvalidParameterValue,
                    exception.NetworkError) as exc:
                LOG.error(
                    "Failed to update port %(port)s for tenant network: "
                    "%(err)s",
                    {'port': port.uuid, 'err': exc}
                )
                raise exception.NetworkError(
                    _("Failed to configure tenant network for port "
                      "%(port)s: %(err)s")
                    % {'port': port.uuid, 'err': exc}
                )

    def _remove_tenant_networks(self, task):
        """Remove tenant networks from a node.

        This method attempts to reset the tenant network configuration for
        each port on the node. It uses the tenant network configuration if
        set, otherwise falls
        back to the port's switchport configuration. If neither is available,
        the port is skipped with a log message.

        :param task: A TaskManager instance.
        """
        LOG.debug("Removing tenant networks from node %(node)s",
                  {'node': task.node.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        errors = []
        for port in task.ports:
            switchport = port.extra.get('switchport') if port.extra else None
            link_info = port.local_link_connection

            if not switchport or not link_info:
                LOG.debug(
                    "Skipping port %(port)s: missing switchport or "
                    "local_link_connection info for tenant network removal.",
                    {'port': port.uuid}
                )
                continue

            config = SwitchPortConfig.from_switchport(switchport)

            if not config or not config.is_valid:
                LOG.debug(
                    "Skipping port %(port)s: missing mode or native_vlan or "
                    "allowed_vlans for tenant network removal.",
                    {'port': port.uuid}
                )
                continue

            try:
                idle_vlan = (idle_config.native_vlan
                             if idle_config else None)
                networking_api.reset_port(
                    task.context,
                    link_info.get('switch_id'),
                    link_info.get('port_id'),
                    config.native_vlan,
                    allowed_vlans=config.allowed_vlans,
                    default_vlan=idle_vlan
                )
            except (exception.InvalidParameterValue,
                    exception.NetworkError) as exc:
                # Accumulate errors to ensure that we attempt to reset each
                # port regardless of if any single port had an error.
                message = (f"Failed to reset tenant network for "
                           f"port {port.uuid}: {exc}")
                LOG.error(message)
                errors.append(message)

        if len(errors) > 0:
            raise exception.NetworkError(
                _("Failed to reset ports for tenant network, "
                  "errors: %(errors)s") %
                {"errors": errors})

    def _add_network(self, task, network_type):
        """Add a network to a node.

        This method attempts to configure the specified network for each port
        on the node. It uses the network configuration if set, otherwise falls
        back to the port's switchport configuration. If neither is available,
        the port is skipped with a log message.

        :param task: A TaskManager instance.
        :param network_type: The type of network to add (e.g.,
                             'provisioning', 'cleaning', 'rescuing',
                             'inspection', 'servicing').
        """
        LOG.info("Adding %(network_type)s network to node %(node)s",
                 {'network_type': network_type, 'node': task.node.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        # Get the config for the network type. It may be
        # overridden by the port's switchport configuration.
        global_config = self._get_switch_port_config(task, network_type)

        for port in task.ports:
            # If the local_link_connection info is missing, skip the port
            link_info = port.local_link_connection
            if not link_info:
                LOG.debug(
                    "Skipping port %(port)s: missing local_link_connection "
                    "info for %(network_type)s network setup.",
                    {'port': port.uuid, 'network_type': network_type}
                )
                continue

            # If the global config is not complete, try to get
            # switchport info from port.extra
            config = global_config
            if config is None or not config.is_valid:
                switchport = (port.extra.get('switchport')
                              if port.extra else None)
                if switchport:
                    config = SwitchPortConfig.from_switchport(switchport)

            if not config or not config.is_valid:
                LOG.debug(
                    "Skipping port %(port)s: missing mode or native_vlan or "
                    "allowed_vlans for %(network_type)s network setup.",
                    {'port': port.uuid, 'network_type': network_type}
                )
                continue

            try:
                idle_vlan = (idle_config.native_vlan
                             if idle_config else None)
                networking_api.update_port(
                    task.context,
                    link_info.get('switch_id'),
                    link_info.get('port_id'),
                    self._get_port_description(port),
                    config.mode,
                    config.native_vlan,
                    allowed_vlans=config.allowed_vlans,
                    lag_name=None,
                    default_vlan=idle_vlan
                )
                LOG.debug(
                    "Configured %(network_type)s network for port %(port)s: "
                    "switch_id=%(switch_id)s, port_name=%(port_name)s, "
                    "config=%(config)s",
                    {
                        'network_type': network_type,
                        'port': port.uuid,
                        'switch_id': link_info.get('switch_id'),
                        'port_name': link_info.get('port_id'),
                        'config': config,
                    }
                )
            except (exception.InvalidParameterValue,
                    exception.NetworkError) as exc:
                LOG.error(
                    "Failed to configure %(network_type)s network for port "
                    "%(port)s: %(err)s",
                    {'network_type': network_type, 'port': port.uuid,
                     'err': exc}
                )
                raise exception.NetworkError(
                    _("Failed to configure %(network_type)s network for port "
                      "%(port)s: %(err)s")
                    % {'network_type': network_type, 'port': port.uuid,
                       'err': exc}
                )

    def _remove_network(self, task, network_type):
        """Remove a network from a node.

        This method attempts to reset the specified network configuration for
        each port
        on the node. It uses the network configuration if set, otherwise falls
        back to the port's switchport configuration. If neither is available,
        the port is skipped with a log message.

        :param task: A TaskManager instance.
        :param network_type: The type of network to remove (e.g.,
                             'provisioning', 'cleaning', 'rescuing',
                             'inspection', 'servicing').
        """
        LOG.info("Removing %(network_type)s network from node %(node)s",
                 {'network_type': network_type, 'node': task.node.uuid})

        # Resolve the idle network configuration (if any)
        idle_config = self._get_switch_port_config(task, network.IDLE_NETWORK)

        # Get the config for the network type. It may be
        # overridden by the port's switchport configuration.
        global_config = self._get_switch_port_config(task, network_type)

        errors = []
        for port in task.ports:
            # Get the switch and port info from the port's
            # local_link_connection
            link_info = port.local_link_connection
            if not link_info:
                LOG.debug(
                    "Skipping port %(port)s: missing local_link_connection "
                    "info for %(network_type)s network removal.",
                    {'port': port.uuid, 'network_type': network_type}
                )
                continue

            # If the global config is not complete, try to get
            # switchport info from port.extra
            config = global_config
            if config is None or not config.is_valid:
                switchport = (port.extra.get('switchport')
                              if port.extra else None)
                if switchport:
                    config = SwitchPortConfig.from_switchport(switchport)

            if not config or not config.is_valid:
                LOG.debug(
                    "Skipping port %(port)s: missing mode or native_vlan or "
                    "allowed_vlans for %(network_type)s network removal.",
                    {'port': port.uuid, 'network_type': network_type}
                )
                continue

            try:
                idle_vlan = (idle_config.native_vlan
                             if idle_config else None)
                networking_api.reset_port(
                    task.context,
                    link_info.get('switch_id'),
                    link_info.get('port_id'),
                    config.native_vlan,
                    allowed_vlans=config.allowed_vlans,
                    default_vlan=idle_vlan
                )
                LOG.debug(
                    "Reset %(network_type)s network for port %(port)s: "
                    "switch_id=%(switch_id)s, port_name=%(port_name)s, "
                    "config=%(config)s",
                    {
                        'network_type': network_type,
                        'port': port.uuid,
                        'switch_id': link_info.get('switch_id'),
                        'port_name': link_info.get('port_id'),
                        'config': config,
                    }
                )
            except (exception.InvalidParameterValue,
                    exception.NetworkError) as exc:
                # Accumulate errors to ensure that we attempt to reset each
                # port regardless of if any single port had an error.
                message = (f"Failed to reset {network_type} network for "
                           f"port {port.uuid}: {exc}")
                LOG.error(message)
                errors.append(message)

        if len(errors) > 0:
            raise exception.NetworkError(
                _("Failed to reset %(network_type)s network, "
                  "errors: %(errors)s") %
                {"network_type": network_type, "errors": errors})

    def add_provisioning_network(self, task):
        """Add the provisioning network to a node.

        This method configures the provisioning network for each port on the
        node by applying the port's switchport configuration with the
        provisioning network VLAN/segment ID override if set.

        :param task: A TaskManager instance.
        """
        self._add_network(task, network.PROVISIONING_NETWORK)

    def remove_provisioning_network(self, task):
        """Remove the provisioning network from a node.

        This method resets the provisioning network configuration for each
        port on the node by restoring the port's default switchport
        configuration.

        :param task: A TaskManager instance.
        """
        self._remove_network(task, network.PROVISIONING_NETWORK)

    def configure_tenant_networks(self, task):
        """Configure tenant networks for a node.

        This method configures tenant network connectivity for each port on
        the node by applying the port's switchport configuration with any
        tenant network overrides.

        :param task: A TaskManager instance.
        """
        self._add_tenant_networks(task)

    def unconfigure_tenant_networks(self, task):
        """Unconfigure tenant networks for a node.

        This method removes tenant network configuration for each port on the
        node by restoring the port's default switchport configuration.

        :param task: A TaskManager instance.
        """
        self._remove_tenant_networks(task)

    def add_cleaning_network(self, task):
        """Add the cleaning network to a node.

        This method configures the cleaning network for each port on the node
        by applying the port's switchport configuration with the cleaning
        network VLAN/segment ID override if set.

        :param task: A TaskManager instance.
        :returns: Empty dictionary as no ports are configured.
        """
        self._add_network(task, network.CLEANING_NETWORK)

    def remove_cleaning_network(self, task):
        """Remove the cleaning network from a node.

        This method resets the cleaning network configuration for each port on
        the node by restoring the port's default switchport configuration.

        :param task: A TaskManager instance.
        """
        self._remove_network(task, network.CLEANING_NETWORK)

    def validate_rescue(self, task):
        """Validate the network interface for rescue operation.

        This method validates that at least one port has the required network
        configuration for rescue operations.

        :param task: A TaskManager instance.
        :raises: InvalidParameterValue if unable to parse network configuration
        """
        self._validate_network_requirements(task, network.RESCUING_NETWORK)

    def add_rescuing_network(self, task):
        """Add the rescuing network to the node.

        This method configures the rescuing network for each port on the node
        by applying the port's switchport configuration with the rescuing
        network VLAN/segment ID override if set.

        :param task: A TaskManager instance.
        :returns: Empty dictionary as no ports are configured.
        """
        self._add_network(task, network.RESCUING_NETWORK)

    def remove_rescuing_network(self, task):
        """Remove the rescuing network from a node.

        This method resets the rescuing network configuration for each port on
        the node by restoring the port's default switchport configuration.

        :param task: A TaskManager instance.
        """
        self._remove_network(task, network.RESCUING_NETWORK)

    def validate_inspection(self, task):
        """Validate that the node has required properties for inspection.

        This method validates that at least one port has the required network
        configuration for inspection operations.

        :param task: A TaskManager instance with the node being checked.
        :raises: InvalidParameterValue if unable to parse network configuration
        """
        self._validate_network_requirements(task, network.INSPECTION_NETWORK)

    def add_inspection_network(self, task):
        """Add the inspection network to the node.

        This method configures the inspection network for each port on the
        node by applying the port's switchport configuration with the
        inspection network VLAN/segment ID override if set.

        :param task: A TaskManager instance.
        :returns: Empty dictionary as no ports are configured.
        """
        self._add_network(task, network.INSPECTION_NETWORK)

    def remove_inspection_network(self, task):
        """Remove the inspection network from a node.

        This method resets the inspection network configuration for each port
        on the node by restoring the port's default switchport configuration.

        :param task: A TaskManager instance.
        """
        self._remove_network(task, network.INSPECTION_NETWORK)

    def need_power_on(self, task):
        """Check if node must be powered on before applying network changes.

        Switch operations can be performed without powering on the node.

        :param task: A TaskManager instance.
        :returns: False as no power state changes are needed.
        """
        return False

    def get_node_network_data(self, task):
        """Return network configuration for node NICs.

        This method returns network configuration data for the node. It first
        checks if static network data is configured on the node itself. If
        present, that takes precedence. Otherwise, it builds network data from
        the ports and portgroups attached to the node.

        The network data is returned in Nova network metadata layout
        (`network_data.json`) format.

        For the ironic-networking interface, this generates:

        - Physical links (type: "phy") for each port with a MAC address
        - VLAN interfaces (type: "vlan") for ports with allowed_vlans
          configured in their switchport settings

        :param task: A TaskManager instance.
        :returns: a dict holding network configuration information adhering
            to Nova network metadata layout (`network_data.json`).
        """
        # Static network data takes precedence
        if task.node.network_data:
            return task.node.network_data

        LOG.debug('Building network data from ports for node %(node)s',
                  {'node': task.node.uuid})

        links = []

        # Process each port to build physical and VLAN links
        for port in task.ports:
            # Skip ports without MAC addresses
            if not port.address:
                LOG.debug('Skipping port %(port)s: no MAC address',
                          {'port': port.uuid})
                continue

            # Create physical link for this port
            phy_link = {
                'id': port.uuid,
                'type': 'phy',
                'ethernet_mac_address': port.address,
                'mtu': 1500
            }
            links.append(phy_link)

            # Check for VLAN configuration in switchport settings
            switchport = port.extra.get('switchport', {}) if port.extra else {}
            allowed_vlans = switchport.get('allowed_vlans', [])

            # Generate VLAN interfaces for each allowed VLAN
            for vlan_id in allowed_vlans:
                vlan_link = {
                    'id': f'{port.uuid}_vlan_{vlan_id}',
                    'type': 'vlan',
                    'vlan_mac_address': port.address,
                    'vlan_id': vlan_id,
                    'vlan_link': port.uuid,
                    'mtu': 1500
                }
                links.append(vlan_link)

        LOG.debug('Generated network data for node %(node)s: %(links)d links',
                  {'node': task.node.uuid, 'links': len(links)})

        # TODO(alegacy): enhance this for LAG support
        return {'links': links}

    def add_servicing_network(self, task):
        """Add the servicing network to the node.

        This method configures the servicing network for each port on the node
        by applying the port's switchport configuration with the servicing
        network VLAN/segment ID override if set.

        :param task: A TaskManager instance.
        :returns: Empty dictionary as no ports are configured.
        """
        self._add_network(task, network.SERVICING_NETWORK)

    def remove_servicing_network(self, task):
        """Remove the servicing network from a node.

        This method resets the servicing network configuration for each port
        on the node by restoring the port's default switchport configuration.

        :param task: A TaskManager instance.
        """
        self._remove_network(task, network.SERVICING_NETWORK)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Networking service manager for Ironic.

The networking service handles network-related operations for Ironic,
providing RPC interfaces for configuring switch ports and network settings.
"""

import functools
import inspect

from oslo_log import log
import oslo_messaging as messaging

from ironic.common import exception
from ironic.common.exception import ConfigInvalid
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import rpc
from ironic.conf import CONF
from ironic.drivers.modules.switch.base import SwitchDriverException
from ironic.drivers.modules.switch.base import SwitchMethodNotImplemented
from ironic.networking import switch_config
from ironic.networking.switch_drivers import driver_adapter
from ironic.networking.switch_drivers import driver_factory
from ironic.networking import utils as networking_utils

LOG = log.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)


def validate_vlan_configuration(
    operation,
    switch_id_arg_name="switch_id",
):
    """Decorator to validate VLAN configuration against allowed/denied lists.

    This decorator extracts native_vlan and allowed_vlans from method
    arguments and validates them against the switch's VLAN configuration.

    :param operation: Description of the operation for error messages.
    :param switch_id_arg_name: Name of the argument containing the switch ID.
                              For multi-switch operations, this should be the
                              argument containing the list of switch IDs.
    :returns: Decorated function that applies VLAN validation.
    """

    def decorator(func):
        # Precompute the function signature at decoration time
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(self, context, *args, **kwargs):
            """Wrapper to validate VLAN config before executing operation."""
            # Use the precomputed signature to bind the arguments
            bound_args = sig.bind(self, context, *args, **kwargs)
            bound_args.apply_defaults()

            # Extract VLAN-related arguments
            native_vlan = bound_args.arguments.get("native_vlan", None)
            allowed_vlans = bound_args.arguments.get("allowed_vlans", [])
            switch_id_value = bound_args.arguments.get(
                switch_id_arg_name, None
            )

            # For multi-switch operations, use the primary (first) switch
            if isinstance(switch_id_value, (list, tuple)) and switch_id_value:
                primary_switch_id = switch_id_value[0]
            else:
                primary_switch_id = switch_id_value

            # Only validate if we have a switch_id and native_vlan
            if primary_switch_id and native_vlan is not None:
                # Get the switch driver
                driver = self._get_switch_driver(primary_switch_id)

                # Build list of VLANs to check
                vlans_to_check = []
                if native_vlan:
                    vlans_to_check.append(native_vlan)
                if allowed_vlans:
                    vlans_to_check.extend(allowed_vlans)

                # Validate VLAN configuration
                switch_config.validate_vlan_configuration(
                    vlans_to_check,
                    driver,
                    primary_switch_id,
                    operation,
                )

            return func(self, context, *args, **kwargs)

        return wrapper

    return decorator


def _get_switch_config_filename():
    return CONF.ironic_networking.driver_config_dir + "/switches.conf"


class NetworkingManager(object):
    """Ironic Networking service manager."""

    # NOTE(alegacy): This must be in sync with rpcapi.NetworkingAPI's.
    RPC_API_VERSION = "1.0"

    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, host, topic=None):
        if not host:
            host = CONF.host
        self.host = host
        if topic is None:
            if networking_utils.rpc_transport() == "json-rpc":
                # Always use the JSON-RPC config for both topic and host
                host_ip = CONF.ironic_networking_json_rpc.host_ip
                port = CONF.ironic_networking_json_rpc.port
                topic_host = f"{host_ip}:{port}"
                topic = f"ironic.{topic_host}"
                self.host = topic_host
            else:
                topic = rpc.NETWORKING_TOPIC
        self.topic = topic

        # Tell the RPC service which json-rpc config group to use for
        # networking. This enables separate listener configuration.
        self.json_rpc_conf_group = "ironic_networking_json_rpc"

        # Initialize driver adapter for configuration preprocessing
        self._driver_adapter = None

        # Initialize switch driver factory (will be set properly in init_host)
        self._switch_driver_factory = None

    def prepare_host(self):
        """Prepare host for networking service initialization.

        This method is called by the RPC service before starting the listener.
        Since networking host configuration is now handled in __init__,
        this method is a no-op.
        """
        pass

    def init_host(self, admin_context=None):
        """Initialize the networking service host.

        This method implements two-phase driver initialization:
        1. Load driver classes (but don't initialize instances yet)
        2. Get translators from driver classes and preprocess config
        3. Initialize driver instances (config files now exist)

        :param admin_context: admin context (unused but kept for compatibility)
        """
        LOG.info("Initializing networking service on host %s", self.host)

        # Phase 1: Load driver classes (invoke_on_load=False)
        self._switch_driver_factory = (
            driver_factory.get_switch_driver_factory()
        )
        available_drivers = self._switch_driver_factory.names
        if not available_drivers:
            LOG.error("No switch drivers loaded")
            raise ConfigInvalid(_("No switch drivers loaded"))

        LOG.info("Available switch drivers: %s", ", ".join(available_drivers))

        # Phase 2: Get driver classes and create adapter
        try:
            # Get driver classes (not instances)
            driver_classes = self._switch_driver_factory.get_driver_classes()

            # Create adapter with driver classes
            self._driver_adapter = (
                driver_adapter.NetworkingDriverAdapter(driver_classes)
            )

            # Preprocess config - writes driver-specific config files
            count = self._driver_adapter.preprocess_config(
                _get_switch_config_filename()
            )
            LOG.info(
                "Generated %d driver-specific config files during init", count
            )
        except Exception as e:
            LOG.exception("Failed to preprocess driver configuration: %s", e)
            raise

        # Phase 3: Now initialize driver instances (config files are ready)
        try:
            self._switch_driver_factory.initialize_drivers()
            LOG.info("Successfully initialized switch driver instances")
        except Exception as e:
            LOG.exception("Failed to initialize switch drivers: %s", e)
            raise


    def _get_switch_driver(self, switch_id):
        """Get the appropriate switch driver for a switch.

        This method finds the correct driver for a switch by checking each
        available driver to see if it is configured to handle the switch.
        If multiple drivers can handle the same switch, the first one found
        is used.

        :param switch_id: Identifier of the switch.
        :returns: Switch driver instance.
        :raises: NetworkError if no drivers are available.
        :raises: SwitchNotFound if no driver supports the switch.
        """
        available_drivers = self._switch_driver_factory.names
        if not available_drivers:
            raise exception.NetworkError(
                _(
                    "No switch drivers are available. Please configure "
                    "enabled_switch_drivers in the networking section."
                )
            )

        # Check each driver to see if it can handle this switch
        for driver_name in available_drivers:
            try:
                driver = self._switch_driver_factory.get_driver(driver_name)

                # Check if this driver is configured for this switch
                if driver.is_switch_configured(switch_id):
                    LOG.debug(
                        "Using switch driver '%s' for switch '%s'",
                        driver_name,
                        switch_id,
                    )
                    return driver

            except exception.DriverNotFound:
                LOG.warning(
                    "Switch driver '%s' not found, skipping", driver_name
                )
                continue
            except Exception as e:
                LOG.warning(
                    "Error checking driver '%s' for switch '%s': %s",
                    driver_name,
                    switch_id,
                    e,
                )
                continue

        # No driver found that supports this switch
        raise exception.SwitchNotFound(switch_id=switch_id)

    def _update_port_impl(
        self,
        switch_id,
        port_name,
        description,
        mode,
        native_vlan,
        allowed_vlans=None,
        default_vlan=None,
        lag_name=None,
    ):
        """Implementation of port update operation.

        :param switch_id: Identifier of the network switch.
        :param port_name: Name of the port to update.
        :param description: Description for the port.
        :param mode: Port mode (e.g., 'access', 'trunk').
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: Allowed VLAN IDs to be removed (optional).
        :param default_vlan: VLAN ID to restore onto the port (optional).
        :param lag_name: Name of the LAG (optional).
        :returns: Dictionary containing the updated port configuration.
        """
        # Get the appropriate switch driver
        driver = self._get_switch_driver(switch_id)

        # Call the driver's update_port method
        driver.update_port(
            switch_id,
            port_name,
            description,
            mode,
            native_vlan,
            allowed_vlans=allowed_vlans,
            default_vlan=default_vlan,
            lag_name=lag_name,
        )

        # Return configuration summary
        port_config = {
            "switch_id": switch_id,
            "port_name": port_name,
            "description": description,
            "mode": mode,
            "native_vlan": native_vlan,
            "allowed_vlans": allowed_vlans or [],
            "default_vlan": default_vlan,
            "lag_name": lag_name,
            "status": "configured",
        }

        LOG.info(
            "Successfully configured port %(port)s on switch %(switch)s",
            {"port": port_name, "switch": switch_id},
        )

        return port_config

    @METRICS.timer("NetworkingManager.update_port")
    @messaging.expected_exceptions(
        exception.InvalidParameterValue,
        exception.NetworkError,
        exception.SwitchNotFound,
    )
    @validate_vlan_configuration("update_port")
    def update_port(
        self,
        context,
        switch_id,
        port_name,
        description,
        mode,
        native_vlan,
        allowed_vlans=None,
        default_vlan=None,
        lag_name=None,
    ):
        """Update a network switch port configuration.

        :param context: request context.
        :param switch_id: Identifier of the network switch.
        :param port_name: Name of the port to update.
        :param description: Description for the port.
        :param mode: Port mode (e.g., 'access', 'trunk').
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: Allowed VLAN IDs to be removed (optional).
        :param default_vlan: VLAN ID to restore onto the port (optional).
        :param lag_name: Name of the LAG if port is part of a link aggregation
                         group (optional).
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the updated port configuration.
        """
        LOG.debug(
            "RPC update_port called for switch %(switch)s, port %(port)s",
            {"switch": switch_id, "port": port_name},
        )

        # Validate mode
        valid_modes = ["access", "trunk"]
        if mode not in valid_modes:
            raise exception.InvalidParameterValue(
                _("mode must be one of: %s") % ", ".join(valid_modes)
            )

        # Validate VLAN ID
        if (
            not isinstance(native_vlan, int)
            or native_vlan < 1
            or native_vlan > 4094
        ):
            raise exception.InvalidParameterValue(
                _("native_vlan must be an integer between 1 and 4094")
            )

        # Validate allowed_vlans if provided
        if allowed_vlans is not None:
            if not isinstance(allowed_vlans, (list, tuple)):
                raise exception.InvalidParameterValue(
                    _("allowed_vlans must be a list or tuple")
                )
            for vlan in allowed_vlans:
                if not isinstance(vlan, int) or vlan < 1 or vlan > 4094:
                    raise exception.InvalidParameterValue(
                        _(
                            "Each VLAN in allowed_vlans must be an integer "
                            "between 1 and 4094"
                        )
                    )

        try:
            return self._update_port_impl(
                switch_id,
                port_name,
                description,
                mode,
                native_vlan,
                allowed_vlans=allowed_vlans,
                default_vlan=default_vlan,
                lag_name=lag_name,
            )
        except exception.InvalidParameterValue:
            # Re-raise validation errors as-is
            raise
        except exception.NetworkError:
            # Re-raise NetworkError as-is
            raise
        except Exception as e:
            LOG.exception(
                "Failed to configure port %(port)s on switch "
                "%(switch)s",
                {"port": port_name, "switch": switch_id},
            )
            raise exception.NetworkError(
                _("Failed to configure network port: %s") % e
            ) from e

    def _reset_port_impl(
        self, switch_id, port_name, native_vlan, allowed_vlans=None,
        default_vlan=None
    ):
        """Implementation of port reset operation.

        :param switch_id: Identifier of the network switch.
        :param port_name: Name of the port to reset.
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: Allowed VLAN IDs to be removed (optional).
        :param default_vlan: VLAN ID to restore onto the port (optional).
        :returns: Dictionary containing the reset port configuration.
        """
        # Get the appropriate switch driver
        driver = self._get_switch_driver(switch_id)

        # Call the driver's reset_port method
        driver.reset_port(
            switch_id, port_name, native_vlan,
            allowed_vlans=allowed_vlans, default_vlan=default_vlan
        )

        # Return reset configuration summary
        port_config = {
            "switch_id": switch_id,
            "port_name": port_name,
            "description": "Default port configuration",
            "mode": "access",
            "native_vlan": native_vlan,
            "allowed_vlans": [],
            "default_vlan": default_vlan,
            "status": "reset",
        }

        LOG.info(
            "Successfully reset port %(port)s on switch %(switch)s",
            {"port": port_name, "switch": switch_id},
        )

        return port_config

    @METRICS.timer("NetworkingManager.reset_port")
    @messaging.expected_exceptions(
        exception.InvalidParameterValue,
        exception.NetworkError,
        exception.SwitchNotFound,
    )
    def reset_port(
        self,
        context,
        switch_id,
        port_name,
        native_vlan,
        allowed_vlans=None,
        default_vlan=None,
    ):
        """Reset a network switch port to default configuration.

        :param context: request context.
        :param switch_id: Identifier of the network switch.
        :param port_name: Name of the port to reset.
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: Allowed VLAN IDs to be removed (optional).
        :param default_vlan: VLAN ID to restore onto the port (optional).
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the reset port configuration.
        """
        LOG.debug(
            "RPC reset_port called for switch %(switch)s, port %(port)s",
            {"switch": switch_id, "port": port_name},
        )

        try:
            return self._reset_port_impl(
                switch_id, port_name, native_vlan,
                allowed_vlans=allowed_vlans, default_vlan=default_vlan
            )
        except exception.InvalidParameterValue:
            # Re-raise validation errors as-is
            raise
        except exception.NetworkError:
            # Re-raise NetworkError as-is
            raise
        except exception.SwitchNotFound:
            # Re-raise SwitchNotFound as-is
            raise
        except Exception as e:
            LOG.exception(
                "Failed to reset port %(port)s on switch "
                "%(switch)s",
                {"port": port_name, "switch": switch_id},
            )
            raise exception.NetworkError(
                _("Failed to reset network port: %s") % e
            ) from e

    @METRICS.timer("NetworkingManager.update_lag")
    @messaging.expected_exceptions(
        exception.InvalidParameterValue,
        exception.NetworkError,
        exception.SwitchNotFound,
        exception.Invalid,
        SwitchMethodNotImplemented,
    )
    @validate_vlan_configuration("update_lag",
                                 switch_id_arg_name="switch_ids")
    def update_lag(
        self,
        context,
        switch_ids,
        lag_name,
        description,
        mode,
        native_vlan,
        aggregation_mode,
        allowed_vlans=None,
        default_vlan=None,
    ):
        """Update a link aggregation group (LAG) configuration.

        :param context: request context.
        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG to update.
        :param description: Description for the LAG.
        :param mode: LAG mode (e.g., 'access', 'trunk').
        :param native_vlan: VLAN ID to be removed from the port.
        :param aggregation_mode: Aggregation mode (e.g., 'lacp', 'static').
        :param allowed_vlans: Allowed VLAN IDs to be removed (optional).
        :param default_vlan: VLAN ID to restore onto the port (optional).
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :raises: SwitchMethodNotImplemented - LAG is not yet supported.
        :returns: Dictionary containing the updated LAG configuration.
        """
        raise SwitchMethodNotImplemented(
            _("LAG operations are not yet supported")
        )

    @METRICS.timer("NetworkingManager.delete_lag")
    @messaging.expected_exceptions(
        exception.InvalidParameterValue,
        exception.NetworkError,
        exception.SwitchNotFound,
        exception.Invalid,
        SwitchMethodNotImplemented,
    )
    def delete_lag(self, context, switch_ids, lag_name):
        """Delete a link aggregation group (LAG) configuration (stub).

        :param context: request context.
        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG to delete.
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :raises: SwitchMethodNotImplemented - LAG is not yet supported.
        :returns: Dictionary containing the deletion status.
        """
        raise SwitchMethodNotImplemented(
            _("LAG operations are not yet supported")
        )

    @METRICS.timer("NetworkingManager.get_switches")
    @messaging.expected_exceptions(exception.NetworkError)
    def get_switches(self, context):
        """Get information about all switches from all drivers.

        :param context: Request context
        :returns: Dictionary of switch_id -> switch_info dictionaries
        :raises: NetworkError if driver operations fail
        """
        switches = {}

        if self._switch_driver_factory is None:
            LOG.debug("Switch driver factory not available")
            return switches

        driver_names = self._switch_driver_factory.names
        if not driver_names:
            LOG.debug("No switch drivers available")
            return switches

        for driver_name in driver_names:
            try:
                driver = self._switch_driver_factory.get_driver(driver_name)
            except (exception.DriverNotFound, exception.DriverLoadError) as e:
                LOG.error(
                    "Error accessing driver %(driver)s: %(error)s",
                    {"driver": driver_name, "error": e}
                )
                raise exception.NetworkError(
                    _("Error accessing driver %(driver)s: %(error)s") % {
                        "driver": driver_name, "error": e
                    }
                ) from e

            try:
                switch_ids = driver.get_switch_ids()
                for switch_id in switch_ids:
                    switch_info = driver.get_switch_info(switch_id)
                    if switch_info:
                        switches[switch_id] = switch_info
            except SwitchDriverException as e:
                LOG.error(
                    "Failed to get switch information from driver "
                    "%(driver)s: %(error)s",
                    {"driver": driver_name, "error": e}
                )
                raise exception.NetworkError(
                    _("Failed to get switch information from driver "
                      "%(driver)s: %(error)s") % {
                        "driver": driver_name, "error": e
                    }
                ) from e

        LOG.info("Successfully retrieved %(count)d switch config sections",
                 {"count": len(switches)})
        return switches

    def cleanup(self):
        """Clean up resources."""
        pass



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import threading

from oslo_service import backend
backend.init_backend(backend.BackendType.THREADING)

from ironic.common import i18n  # noqa

# NOTE(TheJulia): We are setting a default thread stack size for all the
# following thread invocations. Ultimately, while the python minimum is
# any positive number with a minimum of 32768 Bytes, in 4096 Byte
# increments. On some distributions/kernel configs this value can be
# smaller, however, this works on every tested linux distribution.
threading.stack_size(
    int(os.environ.get('IRONIC_THREAD_STACK_SIZE', 131072)))
i18n.install('ironic')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import dataclasses

from ironic.common import exception
from ironic.drivers.modules.network.switchport_config import SwitchPortConfig
from ironic.tests import base as test_base


class TestSwitchPortConfigFromString(test_base.TestCase):

    def test_access_with_native_vlan(self):
        config = SwitchPortConfig.from_string('access/native_vlan=100')
        self.assertEqual('access', config.mode)
        self.assertEqual(100, config.native_vlan)
        self.assertIsNone(config.allowed_vlans)

    def test_trunk_with_native_vlan(self):
        config = SwitchPortConfig.from_string('trunk/native_vlan=200')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(200, config.native_vlan)
        self.assertIsNone(config.allowed_vlans)

    def test_hybrid_with_native_vlan(self):
        config = SwitchPortConfig.from_string('hybrid/native_vlan=300')
        self.assertEqual('hybrid', config.mode)
        self.assertEqual(300, config.native_vlan)

    def test_with_allowed_vlans(self):
        config = SwitchPortConfig.from_string(
            'trunk/native_vlan=100/allowed_vlans=100,200,300')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(100, config.native_vlan)
        self.assertEqual([100, 200, 300], config.allowed_vlans)

    def test_with_allowed_vlans_single(self):
        config = SwitchPortConfig.from_string(
            'access/native_vlan=50/allowed_vlans=50')
        self.assertEqual('access', config.mode)
        self.assertEqual(50, config.native_vlan)
        self.assertEqual([50], config.allowed_vlans)

    def test_with_allowed_vlans_ranges(self):
        config = SwitchPortConfig.from_string(
            'trunk/native_vlan=100/allowed_vlans=1,2,4-7,9')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(100, config.native_vlan)
        self.assertEqual([1, 2, 4, 5, 6, 7, 9],
                         config.allowed_vlans)

    def test_with_allowed_vlans_range_only(self):
        config = SwitchPortConfig.from_string(
            'trunk/native_vlan=10/allowed_vlans=100-103')
        self.assertEqual([100, 101, 102, 103],
                         config.allowed_vlans)

    def test_mode_only_raises(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string, 'access')

    def test_whitespace_stripped(self):
        config = SwitchPortConfig.from_string(
            ' trunk / native_vlan = 100 ')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(100, config.native_vlan)

    def test_empty_string_raises(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string, '')

    def test_invalid_vlan_not_integer(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string, 'access/native_vlan=abc')

    def test_invalid_allowed_vlans_element(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string,
            'trunk/native_vlan=10/allowed_vlans=a,b')

    def test_invalid_allowed_vlans_bad_range(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string,
            'trunk/native_vlan=10/allowed_vlans=1-2-3')

    def test_invalid_mode(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string,
            'bogus/native_vlan=10')

    def test_unknown_key_ignored(self):
        config = SwitchPortConfig.from_string(
            'trunk/native_vlan=10/unknown=42/allowed_vlans=10,20')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(10, config.native_vlan)
        self.assertEqual([10, 20], config.allowed_vlans)

    def test_invalid_missing_equals(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_string,
            'trunk/no_equals')

    def test_network_type_in_error_message(self):
        self.assertRaisesRegex(
            exception.InvalidParameterValue, 'cleaning',
            SwitchPortConfig.from_string,
            'trunk/native_vlan=abc', network_type='cleaning')

    def test_frozen(self):
        config = SwitchPortConfig.from_string(
            'access/native_vlan=10')
        self.assertRaises(
            dataclasses.FrozenInstanceError,
            setattr, config, 'mode', 'trunk')


class TestSwitchPortConfigFromSwitchport(test_base.TestCase):

    def test_full_switchport_dict(self):
        switchport = {
            'mode': 'trunk',
            'native_vlan': 100,
            'allowed_vlans': [100, 200],
        }
        config = SwitchPortConfig.from_switchport(switchport)
        self.assertEqual('trunk', config.mode)
        self.assertEqual(100, config.native_vlan)
        self.assertEqual([100, 200], config.allowed_vlans)

    def test_mode_and_native_vlan_only(self):
        switchport = {'mode': 'access', 'native_vlan': 50}
        config = SwitchPortConfig.from_switchport(switchport)
        self.assertEqual('access', config.mode)
        self.assertEqual(50, config.native_vlan)
        self.assertIsNone(config.allowed_vlans)

    def test_mode_only(self):
        switchport = {'mode': 'access'}
        config = SwitchPortConfig.from_switchport(switchport)
        self.assertEqual('access', config.mode)
        self.assertIsNone(config.native_vlan)
        self.assertFalse(config.is_valid)

    def test_no_mode_returns_none(self):
        switchport = {'native_vlan': 100}
        config = SwitchPortConfig.from_switchport(switchport)
        self.assertIsNone(config)

    def test_empty_dict_returns_none(self):
        config = SwitchPortConfig.from_switchport({})
        self.assertIsNone(config)

    def test_empty_mode_returns_none(self):
        config = SwitchPortConfig.from_switchport({'mode': ''})
        self.assertIsNone(config)

    def test_invalid_mode_raises(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_switchport,
            {'mode': 'bogus', 'native_vlan': 10})

    def test_native_vlan_not_integer_raises(self):
        self.assertRaises(
            exception.InvalidParameterValue,
            SwitchPortConfig.from_switchport,
            {'mode': 'access', 'native_vlan': 'abc'})


class TestSwitchPortConfigIsValid(test_base.TestCase):

    def test_access_with_native_vlan(self):
        config = SwitchPortConfig(mode='access', native_vlan=10)
        self.assertTrue(config.is_valid)

    def test_access_native_vlan_zero(self):
        config = SwitchPortConfig(mode='access', native_vlan=0)
        self.assertTrue(config.is_valid)

    def test_access_missing_native_vlan(self):
        config = SwitchPortConfig(mode='access')
        self.assertFalse(config.is_valid)

    def test_access_with_only_allowed_vlans(self):
        config = SwitchPortConfig(
            mode='access', allowed_vlans=[10])
        self.assertFalse(config.is_valid)

    def test_trunk_with_allowed_vlans(self):
        config = SwitchPortConfig(
            mode='trunk', allowed_vlans=[10, 20])
        self.assertTrue(config.is_valid)

    def test_trunk_missing_allowed_vlans(self):
        config = SwitchPortConfig(mode='trunk', native_vlan=10)
        self.assertFalse(config.is_valid)

    def test_hybrid_with_allowed_vlans(self):
        config = SwitchPortConfig(
            mode='hybrid', allowed_vlans=[10, 20])
        self.assertTrue(config.is_valid)

    def test_hybrid_missing_allowed_vlans(self):
        config = SwitchPortConfig(mode='hybrid', native_vlan=10)
        self.assertFalse(config.is_valid)

    def test_empty_mode(self):
        config = SwitchPortConfig(mode='', native_vlan=10)
        self.assertFalse(config.is_valid)


class TestSwitchPortConfigEquality(test_base.TestCase):

    def test_equal_instances(self):
        a = SwitchPortConfig(mode='access', native_vlan=10)
        b = SwitchPortConfig(mode='access', native_vlan=10)
        self.assertEqual(a, b)

    def test_different_mode(self):
        a = SwitchPortConfig(mode='access', native_vlan=10)
        b = SwitchPortConfig(mode='trunk', native_vlan=10)
        self.assertNotEqual(a, b)

    def test_different_vlan(self):
        a = SwitchPortConfig(mode='access', native_vlan=10)
        b = SwitchPortConfig(mode='access', native_vlan=20)
        self.assertNotEqual(a, b)

    def test_from_string_equals_constructor(self):
        a = SwitchPortConfig.from_string('access/native_vlan=10')
        b = SwitchPortConfig(mode='access', native_vlan=10)
        self.assertEqual(a, b)

    def test_from_switchport_equals_constructor(self):
        a = SwitchPortConfig.from_switchport(
            {'mode': 'trunk', 'native_vlan': 100,
             'allowed_vlans': [100, 200]})
        b = SwitchPortConfig(
            mode='trunk', native_vlan=100, allowed_vlans=[100, 200])
        self.assertEqual(a, b)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for DRAC inspection interface
"""

from unittest import mock

from oslo_utils import units
import sushy

from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.drac import inspect as drac_inspect
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import inspect as redfish_inspect
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.drivers.modules.drac import utils as test_utils
from ironic.tests.unit.objects import utils as obj_utils

INFO_DICT = test_utils.INFO_DICT


class DracRedfishInspectionTestCase(test_utils.BaseDracTest):
    def setUp(self):
        super(DracRedfishInspectionTestCase, self).setUp()
        self.config(enabled_hardware_types=['idrac'],
                    enabled_power_interfaces=['idrac-redfish'],
                    enabled_management_interfaces=['idrac-redfish'],
                    enabled_inspect_interfaces=['idrac-redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='idrac',
            driver_info=INFO_DICT)

    def init_system_mock(self, system_mock, **properties):
        system_mock.reset()
        system_mock.boot.mode = 'uefi'
        system_mock.bios.attributes = {
            'PxeDev1EnDis': 'Enabled', 'PxeDev2EnDis': 'Disabled',
            'PxeDev3EnDis': 'Disabled', 'PxeDev4EnDis': 'Disabled',
            'PxeDev1Interface': 'NIC.Integrated.1-1-1',
            'PxeDev2Interface': None, 'PxeDev3Interface': None,
            'PxeDev4Interface': None}

        system_mock.memory_summary.size_gib = 2

        system_mock.processors.summary = '8', 'MIPS'

        system_mock.simple_storage.disks_sizes_bytes = (
            1 * units.Gi, units.Gi * 3, units.Gi * 5)
        system_mock.storage.volumes_sizes_bytes = (
            2 * units.Gi, units.Gi * 4, units.Gi * 6)

        system_mock.ethernet_interfaces.summary = {
            '00:11:22:33:44:55': sushy.STATE_ENABLED,
            '24:6E:96:70:49:00': sushy.STATE_DISABLED}
        member_data = [{
            'description': 'Integrated NIC 1 Port 1 Partition 1',
            'name': 'System Ethernet Interface',
            'full_duplex': False,
            'identity': 'NIC.Integrated.1-1-1',
            'mac_address': '24:6E:96:70:49:00',
            'mtu_size': None,
            'speed_mbps': 0,
            'vlan': None}]
        system_mock.ethernet_interfaces.get_members.return_value = [
            test_utils.dict_to_namedtuple(values=interface)
            for interface in member_data
        ]
        return system_mock

    def _setup_lldp_system_mock(self, mock_get_system):
        """System mock for LLDP tests."""
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.identity = 'System.Embedded.1'
        return system_mock

    def _setup_dell_connection_mock(self, system_mock, url='https://bmc.example.com/redfish/v1'):
        """Helper to setup Dell connection mock for LLDP tests."""
        mock_conn = mock.MagicMock()
        mock_conn._url = url
        system_mock._conn = mock_conn
        return mock_conn

    def _create_switch_connections_response(self, members):
        """Create a Mock response for Dell switch connections."""
        mock_response = mock.MagicMock()
        mock_response.json.return_value = {'Members': members}
        return mock_response

    def test_get_properties(self):
        expected = redfish_utils.COMMON_PROPERTIES
        driver = drac_inspect.DracRedfishInspect()
        self.assertEqual(expected, driver.get_properties())

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_with_UEFI_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = 'uefi'
        expected_pxe_mac = ['24:6E:96:70:49:00']
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_with_BIOS_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = 'bios'
        mock_manager = mock.MagicMock()
        system_mock.managers = [mock_manager]
        mock_manager_oem = mock_manager.get_oem_extension.return_value
        mock_manager_oem.get_pxe_port_macs_bios.return_value = \
            ['24:6E:96:70:49:00']
        expected_pxe_mac = ['24:6E:96:70:49:00']
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_without_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = None
        expected_pxe_mac = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_inspect.RedfishInspect, 'inspect_hardware',
                       autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_ethernet_interfaces_mac(
            self, mock_get_system, mock_create_ports_if_not_exist,
            mock_inspect_hardware):
        ethernet_interfaces_mac = {'NIC.Integrated.1-1-1':
                                   '24:6E:96:70:49:00'}
        mock_get_system.return_value.sku = None
        mock_inspect_hardware.return_value = states.MANAGEABLE
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_mac_address = mock.Mock()
            task.driver.inspect._get_mac_address.return_value = \
                ethernet_interfaces_mac
            return_value = task.driver.inspect.inspect_hardware(task)
            self.assertEqual(states.MANAGEABLE, return_value)
            mock_create_ports_if_not_exist.assert_called_once_with(
                task, ['24:6E:96:70:49:00'])

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_mac_address_with_ethernet_interfaces(self, mock_get_system):
        self.init_system_mock(mock_get_system.return_value)
        expected_value = {'NIC.Integrated.1-1-1': '24:6E:96:70:49:00'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            return_value = task.driver.inspect._get_mac_address(task)
            self.assertEqual(expected_value, return_value)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_mac_address_without_ethernet_interfaces(self,
                                                          mock_get_system):
        mock_system = self.init_system_mock(mock_get_system.return_value)
        mock_system.ethernet_interfaces.summary = None
        expected_value = {}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            return_value = task.driver.inspect._get_mac_address(task)
            self.assertEqual(expected_value, return_value)
    @mock.patch.object(inspect_utils, 'store_inspection_data', autospec=True)
    @mock.patch.object(inspect_utils, 'get_inspection_data', autospec=True)
    @mock.patch.object(redfish_inspect.RedfishInspect, 'inspect_hardware',
                       autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_sku_as_serial(
            self, mock_get_system, mock_create_ports,
            mock_parent_inspect, mock_get_data, mock_store_data):
        """Test that Dell SKU (service tag) is used as serial_number.

        Dell systems report the motherboard serial in serial_number and the
        actual service tag in SKU. This test verifies that iDRAC inspection
        uses the SKU as the serial_number in the inspection data.
        """
        system_mock = self.init_system_mock(mock_get_system.return_value)
        # Dell reports motherboard serial in serial_number
        system_mock.serial_number = 'MOBO123456'
        # Dell service tag is in SKU - this is what we want as serial
        system_mock.sku = 'DELL-SVC-TAG'

        mock_parent_inspect.return_value = states.MANAGEABLE
        mock_get_data.return_value = {
            'inventory': {
                'system_vendor': {
                    'serial_number': 'MOBO123456',
                    'manufacturer': 'Dell Inc.',
                    'product_name': 'PowerEdge R640'
                }
            },
            'plugin_data': {}
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_mac_address = mock.Mock(return_value={})
            result = task.driver.inspect.inspect_hardware(task)

            self.assertEqual(states.MANAGEABLE, result)
            # Verify store_inspection_data was called with SKU as serial
            mock_store_data.assert_called_once()
            call_args = mock_store_data.call_args
            inventory = call_args[0][1]
            self.assertEqual('DELL-SVC-TAG',
                             inventory['system_vendor']['serial_number'])

    @mock.patch.object(redfish_inspect.RedfishInspect, 'inspect_hardware',
                       autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_no_sku(
            self, mock_get_system, mock_create_ports, mock_parent_inspect):
        """Test inspection when SKU is not available.

        When SKU is not available, the original serial_number from Redfish
        should be preserved.
        """
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.serial_number = 'MOBO123456'
        system_mock.sku = None

        mock_parent_inspect.return_value = states.MANAGEABLE

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_mac_address = mock.Mock(return_value={})
            result = task.driver.inspect.inspect_hardware(task)

            self.assertEqual(states.MANAGEABLE, result)
            # Parent inspect_hardware called but store_inspection_data
            # should not be called again since SKU is None
            mock_parent_inspect.assert_called_once()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_collect_lldp_data_successful_dell_oem(self, mock_get_system):
        """Test successful LLDP data collection from Dell OEM endpoints."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)
        mock_conn = self._setup_dell_connection_mock(system_mock)

        # Mock the HTTP response with switch connections
        members = [
            {
                'FQDD': 'NIC.Integrated.1-1-1',
                'SwitchConnectionID': 'aa:bb:cc:dd:ee:ff',
                'SwitchPortConnectionID': 'Ethernet1/0/1'
            },
            {
                'FQDD': 'NIC.Integrated.1-1-2',
                'SwitchConnectionID': 'aa:bb:cc:dd:ee:gg',
                'SwitchPortConnectionID': 'Ethernet1/8'
            }
        ]
        mock_response = self._create_switch_connections_response(members)
        mock_conn.get.return_value = mock_response

        expected_lldp = {
            'NIC.Integrated.1-1-1': {
                'switch_chassis_id': 'aa:bb:cc:dd:ee:ff',
                'switch_port_id': 'Ethernet1/0/1'
            },
            'NIC.Integrated.1-1-2': {
                'switch_chassis_id': 'aa:bb:cc:dd:ee:gg',
                'switch_port_id': 'Ethernet1/8'
            }
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._collect_lldp_data(task, system_mock)
            self.assertEqual(expected_lldp, result)

    @mock.patch.object(redfish_inspect.RedfishInspect, '_collect_lldp_data',
                       autospec=True)
    @mock.patch.object(drac_inspect.DracRedfishInspect,
                       '_get_dell_switch_connections', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_collect_lldp_data_fallback_to_standard(self, mock_get_system,
                                                    mock_get_connections,
                                                    mock_super_collect):
        """Test fallback to standard Redfish LLDP when Dell OEM fails."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)

        # Mock _get_dell_switch_connections to raise an exception
        mock_get_connections.side_effect = Exception("Dell OEM failed")

        # Mock fallback response
        mock_super_collect.return_value = {
            'NIC.Integrated.1-1-1': {
                'switch_chassis_id': 'fallback_chassis',
                'switch_port_id': 'fallback_port'
            }
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._collect_lldp_data(task, system_mock)
            # Should return the fallback data
            mock_super_collect.assert_called_once_with(
                task.driver.inspect, task, system_mock)
            self.assertEqual(mock_super_collect.return_value, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_collect_lldp_data_filters_no_link(self, mock_get_system):
        """Test that 'No Link' connections are filtered out."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)
        mock_conn = self._setup_dell_connection_mock(system_mock)

        # Mock the HTTP response with mixed valid/invalid connections
        members = [
            {
                'FQDD': 'NIC.Integrated.1-1-1',
                'SwitchConnectionID': 'aa:bb:cc:dd:ee:ff',
                'SwitchPortConnectionID': 'Ethernet1/8'
            },
            {
                'FQDD': 'NIC.Integrated.1-1-2',
                'SwitchConnectionID': 'No Link',
                'SwitchPortConnectionID': 'No Link'
            },
            {
                'FQDD': None,
                'SwitchConnectionID': 'aa:bb:cc:dd:ee:gg',
                'SwitchPortConnectionID': 'Ethernet1/8'
            }
        ]
        mock_response = self._create_switch_connections_response(members)
        mock_conn.get.return_value = mock_response

        expected_lldp = {
            'NIC.Integrated.1-1-1': {
                'switch_chassis_id': 'aa:bb:cc:dd:ee:ff',
                'switch_port_id': 'Ethernet1/8'
            }
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._collect_lldp_data(task, system_mock)
            self.assertEqual(expected_lldp, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_get_dell_switch_connections_success(self, mock_get_system):
        """Test successful retrieval of Dell switch connections."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)
        mock_conn = self._setup_dell_connection_mock(system_mock)

        expected_members = [
            {'FQDD': 'NIC.Integrated.1-1-1',
             'SwitchConnectionID': 'aa:bb:cc:dd:ee:ff'},
            {'FQDD': 'NIC.Integrated.1-1-2',
             'SwitchConnectionID': 'aa:bb:cc:dd:ee:gg'}
        ]
        mock_response = self._create_switch_connections_response(
            expected_members)
        mock_conn.get.return_value = mock_response

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._get_dell_switch_connections(task)
            self.assertEqual(expected_members, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_get_dell_switch_connections_attr_error(self, mock_get_system):
        """Test AttributeError when accessing private attributes."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)

        # Mock missing _conn attribute
        delattr(system_mock, '_conn')

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._get_dell_switch_connections(task)
            self.assertEqual([], result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_get_dell_switch_connections_conn_error(self, mock_get_system):
        """Test handling of connection errors during HTTP request."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)
        mock_conn = self._setup_dell_connection_mock(system_mock)

        # Mock connection failure
        mock_conn.get.side_effect = Exception("HTTP connection failed")

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._get_dell_switch_connections(task)
            self.assertEqual([], result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_get_dell_switch_connections_empty_response(self, mock_get_system):
        """Test handling of empty response from Dell OEM endpoint."""
        system_mock = self._setup_lldp_system_mock(mock_get_system)
        mock_conn = self._setup_dell_connection_mock(system_mock)

        # Mock empty response
        mock_response = self._create_switch_connections_response([])
        mock_conn.get.return_value = mock_response

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            result = task.driver.inspect._get_dell_switch_connections(task)
            self.assertEqual([], result)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
The Ironic Networking Service
"""

import sys

from oslo_config import cfg
from oslo_log import log
from oslo_service import service

from ironic.command import utils as command_utils
from ironic.common import service as ironic_service
from ironic.networking import rpc_service

CONF = cfg.CONF

LOG = log.getLogger(__name__)


def issue_startup_warnings(conf):
    """Issue any startup warnings for the networking service."""
    # Add any networking-specific startup warnings here
    LOG.info("Starting Ironic Networking Service")


def main():
    # NOTE(alegacy): Safeguard to prevent 'ironic.networking.manager'
    # from being imported prior to the configuration options being loaded.
    assert 'ironic.networking.manager' not in sys.modules

    # Parse config file and command line options, then start logging
    ironic_service.prepare_service('ironic_networking', sys.argv)
    ironic_service.ensure_rpc_transport(CONF)

    mgr = rpc_service.NetworkingRPCService(CONF.host,
                                           'ironic.networking.manager',
                                           'NetworkingManager')

    issue_startup_warnings(CONF)

    launcher = service.launch(CONF, mgr, restart_method='mutate')

    # Set override signals.
    command_utils.handle_signal()

    # Start the processes!
    sys.exit(launcher.wait())


if __name__ == '__main__':
    sys.exit(main())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import yaml

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import actions
from ironic.common.inspection_rules import operators
from ironic.common.inspection_rules import utils
from ironic.common.inspection_rules import validation
from ironic.conf import CONF
from ironic import objects


LOG = log.getLogger(__name__)
SENSITIVE_FIELDS = ['password', 'auth_token', 'bmc_password']


def _synthetic_rule_uuid(index):
    return '00000000-0000-0000-0000-%012x' % index


def get_built_in_rules(rules_file):
    """Load built-in inspection rules."""
    built_in_rules = []

    if not rules_file:
        return built_in_rules

    try:
        with open(rules_file, 'r') as f:
            rules_data = yaml.safe_load(f)

        if not isinstance(rules_data, list):
            msg = (
                _("Built-in rules file (%s) should contain a list of rules")
                % rules_file
            )
            LOG.error(msg)
            raise exception.IronicException(msg)

        synthetic_uuid_index = 0
        for rule_data in rules_data:
            try:
                rule_uuid = rule_data.get('uuid')
                if rule_uuid is None:
                    rule_uuid = _synthetic_rule_uuid(synthetic_uuid_index)
                    synthetic_uuid_index += 1
                rule = {
                    'uuid': rule_uuid,
                    'priority': rule_data.get('priority', 0),
                    'description': rule_data.get('description'),
                    'sensitive': rule_data.get('sensitive', False),
                    'phase': rule_data.get('phase', 'main'),
                    'actions': rule_data.get('actions', []),
                    'conditions': rule_data.get('conditions', []),
                }
                validation.validate_rule(rule, built_in=True)
                built_in_rules.append(rule)
            except Exception as e:
                LOG.error(_("Error parsing built-in rule: %s"), e)
                raise
    except FileNotFoundError:
        LOG.error(_("Built-in rules file not found: %s"),
                  rules_file)
        raise
    except yaml.YAMLError as e:
        LOG.error(_("Error parsing YAML in built-in rules file %s: %s"),
                  rules_file, e)
        raise
    except Exception as e:
        LOG.error(_("Error loading built-in rules from %s: %s"),
                  rules_file, e)
        raise

    return built_in_rules


def check_conditions(task, rule, inventory, plugin_data):
    try:
        if not rule.get('conditions', None):
            return True

        for condition in rule['conditions']:
            op, invtd = utils.parse_inverted_operator(
                condition['op'])

            if op not in operators.OPERATORS:
                supported_ops = ', '.join(operators.OPERATORS.keys())
                msg = (_("Unsupported operator: '%(op)s'. Supported "
                         "operators are: %(supported_ops)s.") % {
                             'op': op, 'supported_ops': supported_ops})
                raise ValueError(msg)

            plugin = operators.get_operator(op)
            if condition.get('loop', []):
                result = plugin().check_with_loop(task, condition, inventory,
                                                  plugin_data)
            else:
                result = plugin().check_condition(task, condition, inventory,
                                                  plugin_data)
            if not result:
                LOG.debug("Skipping rule %(rule)s on node %(node)s: "
                          "condition check '%(op)s': '%(args)s' failed ",
                          {'rule': rule['uuid'], 'node': task.node.uuid,
                           'op': condition['op'], 'args': condition['args']})
                return False
        return True

    except Exception as err:
        LOG.error("Error checking condition on node %(node)s: %(err)s.",
                  {'node': task.node.uuid, 'err': err})
        raise


def apply_actions(task, rule, inventory, plugin_data):

    for action in rule['actions']:
        try:
            op = action['op']
            if op not in actions.ACTIONS:
                supported_ops = ', '.join(actions.ACTIONS.keys())
                msg = (_("Unsupported action: '%(op)s'. Supported actions "
                         "are: %(supported_ops)s.") % {
                             'op': op, 'supported_ops': supported_ops})
                raise ValueError(msg)

            plugin = actions.get_action(op)
            if action.get('loop', []):
                plugin().execute_with_loop(task, action, inventory,
                                           plugin_data)
            else:
                plugin().execute_action(task, action, inventory,
                                        plugin_data)
        except exception.IronicException as err:
            LOG.error("Error applying action on node %(node)s: %(err)s.",
                      {'node': task.node.uuid, 'err': err})
            raise
        except Exception as err:
            LOG.exception("Unexpected error applying action on node "
                          "%(node)s: %(err)s.", {'node': task.node.uuid,
                                                 'err': err})
            raise


def _check_rule(task, rule, inventory, plugin_data):
    """Check a single inspection rule's conditions.

    Applies secret masking appropriate for the rule and evaluates its
    conditions.  Returns the masked data for use by the caller when applying
    actions, so that the same masked views are used consistently.

    :param task: a TaskManager instance
    :param rule: a dict representing the inspection rule
    :param inventory: hardware inventory dict
    :param plugin_data: plugin data dict
    :returns: a ``(masked_inventory, masked_plugin_data)`` tuple if conditions
              passed, or ``None`` if conditions were not met
    :raises: exception.HardwareInspectionFailure, exception.IronicException
    """
    mask_secrets = CONF.inspection_rules.mask_secrets
    is_sensitive_rule = rule.get('sensitive', False)

    should_mask = (mask_secrets == 'always'
                   or mask_secrets == 'sensitive' and not is_sensitive_rule)

    masked_inventory = utils.ShallowMaskDict(
        inventory, sensitive_fields=SENSITIVE_FIELDS,
        mask_enabled=should_mask)

    masked_plugin_data = utils.ShallowMaskDict(
        plugin_data, sensitive_fields=SENSITIVE_FIELDS,
        mask_enabled=should_mask)

    if not check_conditions(task, rule, masked_inventory, masked_plugin_data):
        return None

    return masked_inventory, masked_plugin_data


def apply_rules(task, inventory, plugin_data, inspection_phase):
    """Apply inspection rules to a node."""
    node = task.node

    all_rules = objects.InspectionRule.list(
        context=task.context,
        filters={'phase': inspection_phase})

    built_in_rules = get_built_in_rules(CONF.inspection_rules.built_in_rules)
    rules = all_rules + built_in_rules

    if not rules:
        LOG.debug("No inspection rules to apply for phase "
                  "'%(phase)s on node: %(node)s'", {
                      'phase': inspection_phase,
                      'node': node.uuid})
        return

    rules.sort(key=lambda rule: rule.get('priority', 0), reverse=True)
    LOG.debug("Applying %(count)d inspection rules to node %(node)s",
              {'count': len(rules), 'node': node.uuid})

    for rule in rules:
        try:
            result = _check_rule(task, rule, inventory, plugin_data)
            if result is None:
                continue
            masked_inventory, masked_plugin_data = result
            LOG.info("Applying actions for rule %(rule)s to node %(node)s",
                     {'rule': rule['uuid'], 'node': node.uuid})
            apply_actions(task, rule, masked_inventory, masked_plugin_data)
        except exception.HardwareInspectionFailure:
            raise
        except exception.IronicException as e:
            LOG.error(_("Error applying rule %(rule)s to node "
                        "%(node)s: %(error)s"), {'rule': rule['uuid'],
                                                 'node': node.uuid,
                                                 'error': e})
            raise
        except Exception as e:
            msg = ("Failed to apply rule %(rule)s to node %(node)s: "
                   "%(error)s" % {'rule': rule['uuid'], 'node': node.uuid,
                                  'error': e})

            LOG.exception(msg)

            raise exception.IronicException(msg)

    LOG.info("Finished applying inspection rules to node %s", node.uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.conf import CONF
from ironic.tests.base import TestCase


class ValidateConsoleAllowedPortRanges(TestCase):
    def test_success(self):
        CONF.set_override('port_range', '', 'console')
        self.assertEqual([], CONF.console.port_range)
        CONF.set_override('port_range', '1:10', 'console')
        self.assertEqual([range(1, 10)], CONF.console.port_range)
        CONF.set_override('port_range', '1:10,11:12,13:13', 'console')
        self.assertEqual([range(1, 10), range(11, 12), range(13, 13)],
                         CONF.console.port_range)
        CONF.set_override('port_range', '0:65535', 'console')
        self.assertEqual([range(0, 65535)], CONF.console.port_range)

    def test_invalid_elements(self):
        self.assertRaisesRegex(
            ValueError, r'Value should be in <start>:<end> format',
            CONF.set_override, 'port_range', '10', 'console')
        self.assertRaisesRegex(
            ValueError, r'Value should be in <start>:<end> format',
            CONF.set_override, 'port_range', '10:20:30', 'console')

    def test_non_integer(self):
        self.assertRaisesRegex(
            ValueError, r'Port numbers should be integers',
            CONF.set_override, 'port_range', '10:a', 'console')
        self.assertRaisesRegex(
            ValueError, r'Port numbers should be integers',
            CONF.set_override, 'port_range', 'a:10', 'console')
        self.assertRaisesRegex(
            ValueError, r'Port numbers should be integers',
            CONF.set_override, 'port_range', '10:', 'console')
        self.assertRaisesRegex(
            ValueError, r'Port numbers should be integers',
            CONF.set_override, 'port_range', ':10', 'console')

    def test_bad_order(self):
        self.assertRaisesRegex(
            ValueError, r'Start should not be greater than end.',
            CONF.set_override, 'port_range', '11:10', 'console')
        self.assertRaisesRegex(
            ValueError, r'Start should not be greater than end.',
            CONF.set_override, 'port_range', '10:11,21:20', 'console')

    def test_out_of_range(self):
        self.assertRaisesRegex(
            ValueError, r'Port range should be in \[0, 65535\]',
            CONF.set_override, 'port_range', '-1:10', 'console')
        self.assertRaisesRegex(
            ValueError, r'Port range should be in \[0, 65535\]',
            CONF.set_override, 'port_range', '65530:65536', 'console')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ironic.common.trait_based_networking.base as tbn_base

import lark

FILTER_EXPRESSION_GRAMMAR = r"""
filter_expression: single_expression
                 | compound_expression
                 | paren_expression
                 | function_expression

function_expression: function

single_expression: variable_name comparator string_literal

compound_expression: filter_expression ( boolean_operator filter_expression )+

paren_expression: "(" filter_expression ")"

boolean_operator: "&&" -> op_and
                | "||" -> op_or

comparator: "==" -> equality
          | "!=" -> inequality
          | ">=" -> gt_or_eq
          | ">"  -> gt
          | "<=" -> lt_or_eq
          | "<"  -> lt
          | "=~" -> prefix_match

function: "port.is_port"      -> port_is_port
        | "port.is_portgroup" -> port_is_portgroup

string_literal: /\'[A-Za-z0-9_\-\.]*\'/
variable_name: /[a-z]+\.[a-z\_]+/

%import common.WS
%ignore WS
"""

FILTER_EXPRESSION_GRAMMAR_START_RULE = 'filter_expression'

FilterExpressionParser = lark.Lark(FILTER_EXPRESSION_GRAMMAR,
                                   start=FILTER_EXPRESSION_GRAMMAR_START_RULE)


class FilterExpressionTransformer(lark.Transformer):
    def op_and(self, items):
        return tbn_base.Operator.AND

    def op_or(self, items):
        return tbn_base.Operator.OR

    def equality(self, items):
        return tbn_base.Comparator.EQUALITY

    def inequality(self, items):
        return tbn_base.Comparator.INEQUALITY

    def gt_or_eq(self, items):
        return tbn_base.Comparator.GT_OR_EQ

    def gt(self, items):
        return tbn_base.Comparator.GT

    def lt_or_eq(self, items):
        return tbn_base.Comparator.LT_OR_EQ

    def lt(self, items):
        return tbn_base.Comparator.LT

    def prefix_match(self, items):
        return tbn_base.Comparator.PREFIX_MATCH

    def single_expression(self, items):
        return tbn_base.SingleExpression(items[0], items[1], items[2])

    def compound_expression(self, items):
        return tbn_base.CompoundExpression(items[0], items[1], items[2])

    def paren_expression(self, items):
        return tbn_base.ParenExpression(items[0])

    def filter_expression(self, items):
        return tbn_base.FilterExpression(items[0])

    def port_is_port(self, items):
        return tbn_base.Variables.PORT_IS_PORT

    def port_is_portgroup(self, items):
        return tbn_base.Variables.PORT_IS_PORTGROUP

    def function_expression(self, items):
        return tbn_base.FunctionExpression(items[0])

    def variable_name(self, items):
        token = items[0]
        return tbn_base.Variables(token.value)

    def string_literal(self, items):
        token = items[0]
        # Strip ' characters from literal.
        return token.value[1:-1]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.common.trait_based_networking.loader import ConfigLoader

from ironic.tests import base

import tempfile
import time


class TraitBasedNetworkingConfigLoaderTestCase(base.TestCase):
    def test_config_loader(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: bond_ports\n"
                    "      filter: port.vendor == 'vendor_string'\n"
                    "      min_count: 2\n")
        with tempfile.NamedTemporaryFile(
                mode='w',
                dir=self.tmpdir.name,
                delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')
            self.config(enable_trait_based_networking=True, group='conductor')

            cl = ConfigLoader()
            self.assertEqual(len(cl.traits), 1)

    def test_config_loader_refresh(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: bond_ports\n"
                    "      filter: port.vendor == 'vendor_string'\n"
                    "      min_count: 2\n")
        with tempfile.NamedTemporaryFile(
                mode='w',
                dir=self.tmpdir.name,
                delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')
            self.config(enable_trait_based_networking=True, group='conductor')

            cl = ConfigLoader()
            self.assertEqual(len(cl.traits), 1)
            self.assertEqual(cl.traits[0].name, "CUSTOM_TRAIT_NAME")

            # Refresh, nothing should change.
            old_mtime = cl._last_mtime
            cl.refresh()
            self.assertEqual(old_mtime, cl._last_mtime)
            self.assertEqual(len(cl.traits), 1)
            self.assertEqual(cl.traits[0].name, "CUSTOM_TRAIT_NAME")

            # Try to guarantee the mtime will change.
            time.sleep(1)

            with open(tmpfile.name, mode='w') as newfile:
                contents = ("CUSTOM_TRAIT_NAME_CHANGED:\n"
                            "  actions:\n"
                            "    - action: bond_ports\n"
                            "      filter: port.vendor == 'vendor_string'\n"
                            "      min_count: 2\n")
                newfile.write(contents)
                newfile.close()

                cl.refresh()
                self.assertNotEqual(old_mtime, cl._last_mtime)
                self.assertEqual(len(cl.traits), 1)
                self.assertEqual(cl.traits[0].name,
                                 "CUSTOM_TRAIT_NAME_CHANGED")



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for DRAC RAID interface
"""

from unittest import mock

import sushy
import tenacity

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.drac import raid as drac_raid
from ironic.drivers.modules.drac import utils as drac_utils
from ironic.drivers.modules.redfish import raid as redfish_raid
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.drivers.modules.drac import utils as test_utils
from ironic.tests.unit.objects import utils as obj_utils

from sushy.oem import dell as sushy_oem_idrac

INFO_DICT = test_utils.INFO_DICT


class DracCreateRaidConfigurationHelpersTestCase(test_utils.BaseDracTest):

    def setUp(self):
        super(DracCreateRaidConfigurationHelpersTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='idrac',
                                               driver_info=INFO_DICT)

        self.physical_disk = {
            'id': 'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            'description': ('Disk 1 in Backplane 1 of '
                            'Integrated RAID Controller 1'),
            'controller': 'RAID.Integrated.1-1',
            'manufacturer': 'SEAGATE',
            'model': 'ST600MM0006',
            'media_type': 'hdd',
            'interface_type': 'sas',
            'size_mb': 571776,
            'free_size_mb': 571776,
            'serial_number': 'S0M3EY2Z',
            'firmware_version': 'LS0A',
            'status': 'ok',
            'raid_status': 'ready',
            'sas_address': '500056B37789ABE3',
            'device_protocol': None}

        self.physical_disks = []
        for i in range(8):
            disk = self.physical_disk.copy()
            disk['id'] = ('Disk.Bay.%s:Enclosure.Internal.0-1:'
                          'RAID.Integrated.1-1' % i)
            disk['serial_number'] = 'serial%s' % i

            self.physical_disks.append(disk)

        self.root_logical_disk = {
            'size_gb': 50,
            'raid_level': '1',
            'disk_type': 'hdd',
            'interface_type': 'sas',
            'volume_name': 'root_volume',
            'is_root_volume': True
        }
        self.nonroot_logical_disks = [
            {'size_gb': 100,
             'raid_level': '5',
             'disk_type': 'hdd',
             'interface_type': 'sas',
             'volume_name': 'data_volume1'},
            {'size_gb': 100,
             'raid_level': '5',
             'disk_type': 'hdd',
             'interface_type': 'sas',
             'volume_name': 'data_volume2'}
        ]

        self.logical_disks = (
            [self.root_logical_disk] + self.nonroot_logical_disks)
        self.target_raid_configuration = {'logical_disks': self.logical_disks}
        self.node.target_raid_config = self.target_raid_configuration
        self.node.save()


class DracRedfishRAIDTestCase(test_utils.BaseDracTest):

    def setUp(self):
        super(DracRedfishRAIDTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='idrac',
                                               driver_info=INFO_DICT)
        self.raid = drac_raid.DracRedfishRAID()

    @mock.patch.object(drac_raid, '_wait_till_realtime_ready', autospec=True)
    @mock.patch.object(redfish_raid.RedfishRAID, 'create_configuration',
                       autospec=True)
    def test_create_configuration(self, mock_redfish_create, mock_wait):
        task = mock.Mock(node=self.node, context=self.context)

        self.raid.create_configuration(task)

        mock_wait.assert_called_once_with(task)
        mock_redfish_create.assert_called_once_with(
            self.raid, task, True, True, False)

    @mock.patch.object(drac_raid, '_wait_till_realtime_ready', autospec=True)
    @mock.patch.object(redfish_raid.RedfishRAID, 'delete_configuration',
                       autospec=True)
    def test_delete_configuration(self, mock_redfish_delete, mock_wait):
        task = mock.Mock(node=self.node, context=self.context)

        self.raid.delete_configuration(task)

        mock_wait.assert_called_once_with(task)
        mock_redfish_delete.assert_called_once_with(self.raid, task)

    @mock.patch.object(drac_raid, '_retry_till_realtime_ready', autospec=True)
    def test__wait_till_realtime_ready(self, mock_ready):
        self.node.set_driver_internal_info('cleaning_disable_ramdisk', True)
        self.node.provision_state = states.CLEANING
        self.node.save()
        task = mock.Mock(node=self.node, context=self.context)
        task.driver.power.get_power_state.return_value = states.POWER_OFF
        drac_raid._wait_till_realtime_ready(task)
        task.driver.power.set_power_state.assert_called_once_with(
            task, states.POWER_ON)
        mock_ready.assert_called_once_with(task)

    @mock.patch.object(drac_raid, 'LOG', autospec=True)
    @mock.patch.object(drac_raid, '_retry_till_realtime_ready', autospec=True)
    def test__wait_till_realtime_ready_retryerror(self, mock_ready, mock_log):
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = tenacity.RetryError(3)
        drac_raid._wait_till_realtime_ready(task)
        mock_ready.assert_called_once_with(task)
        self.assertEqual(mock_log.debug.call_count, 1)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready_retry_exceeded(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        drac_raid._retry_till_realtime_ready.retry.stop =\
            tenacity.stop_after_attempt(3)
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.return_value = False

        self.assertRaises(
            tenacity.RetryError,
            drac_raid._retry_till_realtime_ready, task)

        self.assertEqual(3, mock_ready.call_count)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready_retry_fails(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        drac_raid._retry_till_realtime_ready.retry.stop =\
            tenacity.stop_after_attempt(3)
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = [False, exception.RedfishError]

        self.assertRaises(
            exception.RedfishError,
            drac_raid._retry_till_realtime_ready, task)

        self.assertEqual(2, mock_ready.call_count)

    @mock.patch.object(drac_raid, '_is_realtime_ready', autospec=True)
    def test__retry_till_realtime_ready(self, mock_ready):
        drac_raid._retry_till_realtime_ready.retry.sleep = mock.Mock()
        task = mock.Mock(node=self.node, context=self.context)
        mock_ready.side_effect = [False, True]

        is_ready = drac_raid._retry_till_realtime_ready(task)

        self.assertTrue(is_ready)
        self.assertEqual(2, mock_ready.call_count)

    @mock.patch.object(drac_utils, 'LOG', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__is_realtime_ready(self, mock_get_system, mock_log):
        task = mock.Mock(node=self.node, context=self.context)
        fake_manager_oem1 = mock.Mock()
        fake_manager_oem1.lifecycle_service.is_realtime_ready.side_effect = (
            sushy.exceptions.SushyError)
        fake_manager1 = mock.Mock()
        fake_manager1.get_oem_extension.return_value = fake_manager_oem1
        fake_manager_oem2 = mock.Mock()
        fake_manager_oem2.lifecycle_service.is_realtime_ready.return_value = (
            True)
        fake_manager2 = mock.Mock()
        fake_manager2.get_oem_extension.return_value = fake_manager_oem2
        fake_system = mock.Mock(managers=[fake_manager1, fake_manager2])
        mock_get_system.return_value = fake_system

        is_ready = drac_raid._is_realtime_ready(task)

        self.assertTrue(is_ready)
        self.assertEqual(mock_log.debug.call_count, 1)

    def test_validate_correct_vendor(self):
        task = mock.Mock(node=self.node, context=self.context)
        self.node.properties['vendor'] = 'Dell Inc.'
        self.raid.validate(task)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_pre_create_configuration(self, mock_get_system):
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        fake_oem_system = mock.Mock()
        fake_oem_system.change_physical_disk_state.return_value = [
            mock_task_mon1, mock_task_mon2]
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        mock_drive1 = mock.Mock(
            identity='Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1')
        mock_drive2 = mock.Mock(
            identity='Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            capacity_bytes=599550590976)  # mocked size in RAID mode
        mock_drive3 = mock.Mock(
            identity='Disk.Direct.0-0:AHCI.Slot.2-1')

        mock_controller1 = mock.Mock()
        mock_storage1 = mock.Mock(controllers=[mock_controller1],
                                  drives=[mock_drive1, mock_drive2],
                                  identity='RAID.Integrated.1-1')
        mock_storage1.controllers = mock.MagicMock()
        mock_storage1.controllers.get_members.return_value = [
            mock_controller1]

        mock_controller2 = mock.Mock()
        mock_storage2 = mock.Mock(controllers=[mock_controller2],
                                  drives=[mock_drive3],
                                  identity='AHCI.Slot.2-1')
        mock_storage2.controllers = mock.MagicMock()
        mock_storage2.controllers.get_members.return_value = [
            mock_controller2]

        fake_system.storage.get_members.return_value = [
            mock_storage1, mock_storage2]

        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)

        logical_disks_to_create = [{
            'raid_level': '0',
            'size_bytes': 600087461888,  # before RAID conversion
            'physical_disks': [
                'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
            'span_depth': 1,
            'span_length': 1.0,
            'controller': 'RAID.Integrated.1-1'}]

        result = self.raid.pre_create_configuration(
            task, logical_disks_to_create)

        self.assertEqual(
            [{'controller': 'RAID.Integrated.1-1',
              'physical_disks': [
                  'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
              'raid_level': '0',
              'size_bytes': 599550590976,  # recalculated after RAID conversion
              'span_depth': 1,
              'span_length': 1.0}], result)
        fake_oem_system.change_physical_disk_state.assert_called_once_with(
            sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID,
            {mock_controller1: [mock_drive2]})
        mock_task_mon1.wait.assert_called_once_with(CONF.drac.raid_job_timeout)
        mock_task_mon2.wait.assert_not_called()

    def test__get_storage_controller_invalid_identity(self):
        fake_system = mock.Mock()

        mock_storage1 = mock.Mock(storage_controllers=[mock.Mock()],
                                  identity='RAID.Integrated.1-1')
        mock_storage2 = mock.Mock(storage_controllers=[mock.Mock()],
                                  identity='AHCI.Slot.2-1')

        fake_system.storage.get_members.return_value = [
            mock_storage1, mock_storage2]

        self.assertRaises(
            exception.IronicException,
            drac_raid.DracRedfishRAID._get_storage_controller,
            fake_system, 'NonExisting')

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    def test__change_physical_disk_state_attribute_error(self, mock_log):
        fake_oem_system = mock.Mock(spec=[])
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        result = drac_raid.DracRedfishRAID._change_physical_disk_state(
            fake_system, sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID
        )

        self.assertEqual(False, result)
        mock_log.assert_called_once()

    @mock.patch.object(deploy_utils, 'reboot_to_finish_step',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode(
            self, mock_get_system, mock_reboot):
        mock_task_mon = mock.Mock(task_monitor_uri='/TaskService/1')
        mock_oem_controller = mock.Mock()
        mock_oem_controller.convert_to_raid.return_value = mock_task_mon
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.return_value = mock_oem_controller
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertEqual(
                ['/TaskService/1'],
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertEqual(mock_reboot.return_value, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_no_conversion(
            self, mock_get_system):
        mock_oem_controller = mock.Mock()
        mock_oem_controller.convert_to_raid.return_value = None
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.return_value = mock_oem_controller
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_not_raid(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(storage_controllers=None)
        mock_controllers = mock.PropertyMock(
            side_effect=sushy.exceptions.MissingAttributeError)
        type(mock_storage).controllers = mock_controllers
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_not_called()

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_idrac(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(storage_controllers=mock.Mock())
        mock_controllers = mock.PropertyMock(
            side_effect=sushy.exceptions.MissingAttributeError)
        type(mock_storage).controllers = mock_controllers
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_called_once()

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_sushy(
            self, mock_get_system, mock_log):
        mock_storage = mock.Mock(spec=[])
        mock_storage.identity = "Storage 1"
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)
            mock_log.assert_called_once()

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__convert_controller_to_raid_mode_old_sushy_oem(
            self, mock_get_system):
        mock_controller = mock.Mock()
        mock_controller.get_oem_extension.side_effect =\
            sushy.exceptions.ExtensionError
        mock_controllers_col = mock.Mock()
        mock_controllers_col.get_members.return_value = [mock_controller]
        mock_storage = mock.Mock(controllers=mock_controllers_col)
        mock_storage_col = mock.Mock()
        mock_storage_col.get_members.return_value = [mock_storage]
        mock_system = mock.Mock(storage=mock_storage_col)
        mock_get_system.return_value = mock_system

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = self.raid._convert_controller_to_raid_mode(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(result)

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(deploy_utils, 'get_async_step_return_state',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_foreign_async(
            self, mock_get_system, mock_build_agent_options,
            mock_get_async_step_return_state, mock_convert):
        fake_oem_system = mock.Mock()
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system
        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()
        mock_return_state2 = mock.Mock()
        mock_get_async_step_return_state.return_value = mock_return_state2
        mock_oem_task1 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task1 = mock.Mock()
        mock_task1.get_oem_extension.return_value = mock_oem_task1
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon1.task_monitor_uri = '/TaskService/1'
        mock_task_mon1.get_task.return_value = mock_task1
        mock_oem_task2 = mock.Mock(job_type=sushy_oem_idrac.JOB_TYPE_RAID_CONF)
        mock_task2 = mock.Mock()
        mock_task2.get_oem_extension.return_value = mock_oem_task2
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        mock_task_mon2.task_monitor_uri = '/TaskService/2'
        mock_task_mon2.get_task.return_value = mock_task2
        fake_oem_system.clear_foreign_config.return_value = [
            mock_task_mon1, mock_task_mon2]

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(result, mock_return_state2)
        fake_oem_system.clear_foreign_config.assert_called_once()
        mock_build_agent_options.assert_called_once_with(task.node)
        mock_get_async_step_return_state.assert_called_once_with(task.node)
        mock_task_mon1.wait.assert_not_called()
        mock_task_mon2.wait.assert_not_called()
        mock_convert.assert_not_called()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_foreign_sync(
            self, mock_get_system, mock_convert):
        fake_oem_system = mock.Mock()
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system
        mock_get_system.return_value = fake_system
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()
        mock_oem_task1 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task1 = mock.Mock()
        mock_task1.get_oem_extension.return_value = mock_oem_task1
        mock_task_mon1 = mock.Mock(check_is_processing=True)
        mock_task_mon1.get_task.return_value = mock_task1
        mock_oem_task2 = mock.Mock(
            job_type=sushy_oem_idrac.JOB_TYPE_RT_NO_REBOOT_CONF)
        mock_task2 = mock.Mock()
        mock_task2.get_oem_extension.return_value = mock_oem_task2
        mock_task_mon2 = mock.Mock(check_is_processing=False)
        mock_task_mon2.get_task.return_value = mock_task2
        fake_oem_system.clear_foreign_config.return_value = [
            mock_task_mon1, mock_task_mon2]
        mock_convert_state = mock.Mock()
        mock_convert.return_value = mock_convert_state

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(result, mock_convert_state)
        fake_oem_system.clear_foreign_config.assert_called_once()
        mock_task_mon1.wait.assert_called_once_with(CONF.drac.raid_job_timeout)
        mock_task_mon2.wait.assert_not_called()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_clear_foreign_config', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_post_delete_configuration_no_subtasks(
            self, mock_get_system, mock_foreign, mock_convert):
        mock_foreign.return_value = False
        mock_convert.return_value = None
        task = mock.Mock(node=self.node, context=self.context)
        mock_return_state1 = mock.Mock()

        result = self.raid.post_delete_configuration(
            task, None, return_state=mock_return_state1)

        self.assertEqual(mock_return_state1, result)

    @mock.patch.object(drac_raid.LOG, 'warning', autospec=True)
    def test__clear_foreign_config_attribute_error(self, mock_log):
        fake_oem_system = mock.Mock(spec=[])
        fake_system = mock.Mock()
        fake_system.get_oem_extension.return_value = fake_oem_system

        result = drac_raid.DracRedfishRAID._clear_foreign_config(
            fake_system, mock.Mock())

        self.assertEqual(False, result)
        mock_log.assert_called_once()

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test__query_raid_tasks_status(self, mock_acquire):
        driver_internal_info = {'raid_task_monitor_uris': ['/TaskService/123']}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        mock_manager = mock.Mock()
        node_list = [(self.node.uuid, 'idrac', '', driver_internal_info)]
        mock_manager.iter_nodes.return_value = node_list
        task = mock.Mock(node=self.node,
                         driver=mock.Mock(raid=self.raid))
        mock_acquire.return_value = mock.MagicMock(
            __enter__=mock.MagicMock(return_value=task))
        self.raid._check_raid_tasks_status = mock.Mock()

        self.raid._query_raid_tasks_status(mock_manager, self.context)

        self.raid._check_raid_tasks_status.assert_called_once_with(
            task, ['/TaskService/123'])

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test__query_raid_tasks_status_no_task_monitor_url(self, mock_acquire):
        driver_internal_info = {'something': 'else'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        mock_manager = mock.Mock()
        node_list = [(self.node.uuid, 'idrac', '', driver_internal_info)]
        mock_manager.iter_nodes.return_value = node_list
        task = mock.Mock(node=self.node,
                         driver=mock.Mock(raid=self.raid))
        mock_acquire.return_value = mock.MagicMock(
            __enter__=mock.MagicMock(return_value=task))
        self.raid._check_raid_tasks_status = mock.Mock()

        self.raid._query_raid_tasks_status(mock_manager, self.context)

        self.raid._check_raid_tasks_status.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status(self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config done'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123'])

            self.raid._set_success.assert_called_once_with(task)
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_task_still_processing(
            self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config done'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_task_monitor2 = mock.Mock()
        mock_task_monitor2.is_processing = True
        mock_get_task_monitor.side_effect = [
            mock_task_monitor, mock_task_monitor2]

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()
        self.raid._substep_change_physical_disk_state_nonraid = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123', '/TaskService/456'])

            (self.raid._substep_change_physical_disk_state_nonraid
                .assert_not_called())
            self.raid._set_success.assert_not_called()
            self.assertEqual(
                ['/TaskService/456'],
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_not_called()

    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_task_failed(self, mock_get_task_monitor):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/123'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_message = mock.Mock()
        mock_message.message = 'Clear foreign config failed'
        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = 'Failed'
        mock_config_task.messages = [mock_message]
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()
        self.raid._substep_change_physical_disk_state_nonraid = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/123'])

            (self.raid._substep_change_physical_disk_state_nonraid
                .assert_not_called())
            self.raid._set_success.assert_not_called()
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.raid._set_failed.assert_called_once()

    @mock.patch.object(drac_raid.DracRedfishRAID,
                       '_convert_controller_to_raid_mode', autospec=True)
    @mock.patch.object(redfish_utils, 'get_task_monitor', autospec=True)
    def test__check_raid_tasks_status_convert_controller(
            self, mock_get_task_monitor, mock_convert):
        driver_internal_info = {
            'raid_task_monitor_uris': '/TaskService/1',
            'raid_config_substep': 'clear_foreign_config'}
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        mock_config_task = mock.Mock()
        mock_config_task.task_state = sushy.TASK_STATE_COMPLETED
        mock_config_task.task_status = sushy.HEALTH_OK
        mock_task_monitor = mock.Mock()
        mock_task_monitor.is_processing = False
        mock_task_monitor.get_task.return_value = mock_config_task
        mock_get_task_monitor.return_value = mock_task_monitor

        self.raid._set_success = mock.Mock()
        self.raid._set_failed = mock.Mock()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._check_raid_tasks_status(
                task, ['/TaskService/1'])

            mock_convert.assert_called_once_with(task)
            self.raid._set_success.assert_not_called()
            self.raid._set_failed.assert_not_called()
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_task_monitor_uris'))
            self.assertIsNone(
                task.node.driver_internal_info.get('raid_config_substep'))

    @mock.patch.object(manager_utils, 'notify_conductor_resume_deploy',
                       autospec=True)
    @mock.patch.object(manager_utils, 'notify_conductor_resume_clean',
                       autospec=True)
    def test__set_success_clean(self, mock_notify_clean, mock_notify_deploy):
        self.node.clean_step = {'test': 'value'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_success(task)

            mock_notify_clean.assert_called_once_with(task)

    @mock.patch.object(manager_utils, 'notify_conductor_resume_deploy',
                       autospec=True)
    @mock.patch.object(manager_utils, 'notify_conductor_resume_clean',
                       autospec=True)
    def test__set_success_deploy(self, mock_notify_clean, mock_notify_deploy):
        self.node.clean_step = None
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_success(task)

            mock_notify_deploy.assert_called_once_with(task)

    @mock.patch.object(manager_utils, 'deploying_error_handler',
                       autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test__set_failed_clean(self, mock_clean_handler, mock_deploy_handler):
        self.node.clean_step = {'test': 'value'}
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_failed(task, 'error', 'log message')

            mock_clean_handler.assert_called_once_with(
                task, 'error', 'log message')

    @mock.patch.object(manager_utils, 'deploying_error_handler',
                       autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test__set_failed_deploy(self, mock_clean_handler, mock_deploy_handler):
        self.node.clean_step = None
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.raid._set_failed(task, 'error', 'log message')

            mock_deploy_handler.assert_called_once_with(
                task, 'error', 'log message')



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Mapping of hardware health states.

These states represent the overall health condition of the server hardware,
as reported by the BMC. Drivers should convert vendor-specific health values
to these Ironic-level constants.
"""

import enum


class HealthState(enum.Enum):
    """Hardware health states reported by BMC."""

    OK = 'OK'
    """Hardware is functioning normally with no issues detected."""

    WARNING = 'Warning'
    """Hardware has non-critical issues that may require attention."""

    CRITICAL = 'Critical'
    """Hardware has critical issues requiring immediate attention."""
